# Copyright (c) 2014, Cristian Guarnizo.
# Licensed under the BSD 3-clause license (see LICENSE.txt)

import numpy as np
from scipy.special import wofz
from .kern import Kern
from ...core.parameterization import Param
from paramz.transformations import Logexp
from paramz.caching import Cache_this

class EQ_ODE2(Kern):
    """
    Covariance function for second order differential equation driven by an exponentiated quadratic covariance.

    This outputs of this kernel have the form
    .. math::
       \frac{\text{d}^2y_j(t)}{\text{d}^2t} + C_j\frac{\text{d}y_j(t)}{\text{d}t} + B_jy_j(t) = \sum_{i=1}^R w_{j,i} u_i(t)

    where :math:`R` is the rank of the system, :math:`w_{j,i}` is the sensitivity of the :math:`j`th output to the :math:`i`th latent function, :math:`d_j` is the decay rate of the :math:`j`th output and :math:`f_i(t)` and :math:`g_i(t)` are independent latent Gaussian processes goverened by an exponentiated quadratic covariance.

    :param output_dim: number of outputs driven by latent function.
    :type output_dim: int
    :param W: sensitivities of each output to the latent driving function.
    :type W: ndarray (output_dim x rank).
    :param rank: If rank is greater than 1 then there are assumed to be a total of rank latent forces independently driving the system, each with identical covariance.
    :type rank: int
    :param C: damper constant for the second order system.
    :type C: array of length output_dim.
    :param B: spring constant for the second order system.
    :type B: array of length output_dim.

    """
    #This code will only work for the sparseGP model, due to limitations in models for this kernel
    def __init__(self, input_dim=2, output_dim=1, rank=1, W=None, lengthscale=None, C=None, B=None, active_dims=None, name='eq_ode2'):
        #input_dim should be 1, but kern._slice_X is not returning index information required to evaluate kernels        
        assert input_dim == 2, "only defined for 1 input dims"
        super(EQ_ODE2, self).__init__(input_dim=input_dim, active_dims=active_dims, name=name)
        self.rank = rank
        self.output_dim = output_dim

        if lengthscale is None:
            lengthscale = .5+np.random.rand(self.rank)
        else:
            lengthscale = np.asarray(lengthscale)
            assert lengthscale.size in [1, self.rank], "Bad number of lengthscales"
            if lengthscale.size != self.rank:
                lengthscale = np.ones(self.rank)*lengthscale

        if W is None:
            #W = 0.5*np.random.randn(self.output_dim, self.rank)/np.sqrt(self.rank)
            W = np.ones((self.output_dim, self.rank))
        else:
            assert W.shape == (self.output_dim, self.rank)

        if C is None:
            C = np.ones(self.output_dim)

        if B is None:
            B = np.ones(self.output_dim)

        self.C = Param('C', C, Logexp())
        self.B = Param('B', B, Logexp())
        self.lengthscale = Param('lengthscale', lengthscale, Logexp())
        self.W = Param('W', W)
        self.link_parameters(self.lengthscale, self.C, self.B, self.W)

    @Cache_this(limit=3)
    def K(self, X, X2=None):
        #This way is not working, indexes are lost after using k._slice_X
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        index = np.int_(np.round(X[:, 1]))
        index = index.reshape(index.size,)
        X_flag = index[0] >= self.output_dim
        if X2 is None:
            if X_flag:
                #Calculate covariance function for the latent functions
                index -= self.output_dim
                return self._Kuu(X, index)
            else: #Kff full
                raise NotImplementedError
        else:
            #This way is not working, indexes are lost after using k._slice_X
            #index2 = np.asarray(X2, dtype=np.int)
            #index2 = index2.reshape(index2.size,)
            if hasattr(X2, 'values'):
                X2 = X2.values
            index2 = np.int_(np.round(X2[:, 1]))
            index2 = index2.reshape(index2.size,)
            X2_flag = index2[0] >= self.output_dim
            #Calculate cross-covariance function
            if not X_flag and X2_flag:
                index2 -= self.output_dim
                return self._Kfu(X, index, X2, index2) #Kfu
            elif X_flag and not X2_flag:
                index -= self.output_dim
                return self._Kfu(X2, index2, X, index).T #Kuf
            elif X_flag and X2_flag:
                index -= self.output_dim
                index2 -= self.output_dim
                return self._Kusu(X, index, X2, index2) #Ku_s u
            else:
                raise NotImplementedError #Kf_s f

    #Calculate the covariance function for diag(Kff(X,X))
    def Kdiag(self, X):
        if hasattr(X, 'values'):
            index = np.int_(np.round(X[:, 1].values))
        else:
            index = np.int_(np.round(X[:, 1]))
        index = index.reshape(index.size,)
        X_flag = index[0] >= self.output_dim
        
        if X_flag: #Kuudiag        
            return np.ones(X[:,0].shape)
        else: #Kffdiag
            kdiag = self._Kdiag(X)
            return np.sum(kdiag, axis=1)

    #Calculate the covariance function for diag(Kff(X,X))
    def _Kdiag(self, X):
        #This way is not working, indexes are lost after using k._slice_X
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        index = np.int_(X[:, 1])
        index = index.reshape(index.size,)
        
        #terms that move along t
        t = X[:, 0].reshape(X.shape[0], 1)
        d = np.unique(index) #Output Indexes
        B = self.B.values[d]
        C = self.C.values[d]
        S = self.W.values[d, :]
        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]
        #Check where wd becomes complex
        wbool = C*C >= 4.*B
        B = B.reshape(B.size, 1)
        C = C.reshape(C.size, 1)
        alpha = .5*C
        C2 = C*C

        wbool2 = wbool[index]
        ind2t = np.where(wbool2)
        ind3t = np.where(np.logical_not(wbool2))

        #Terms that move along q
        lq = self.lengthscale.values.reshape(1, self.lengthscale.size)
        S2 = S*S
        kdiag = np.empty((t.size, lq.size))

        indD = np.arange(B.size)
        #(1) When wd is real
        if np.any(np.logical_not(wbool)):
            #Indexes of index and t related to (2)
            t1 = t[ind3t]
            ind = index[ind3t]
            d = np.asarray(np.where(np.logical_not(wbool))[0]) #Selection of outputs
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            S2lq = S2[d]*(.5*lq)
            c0 = S2lq*np.sqrt(np.pi)
            w = .5*np.sqrt(4.*B[d] - C2[d])
            alphad = alpha[d]
            w2 = w*w
            gam = alphad + 1j*w
            gamc = alphad - 1j*w
            c1 = .5/(alphad*w2)
            c2 = .5/(gam*w2)
            c = c1 - c2
            #DxQ terms
            nu = lq*(gam*.5)
            K01 = c0*c
            #Nx1 terms
            gamt = -gam[ind]*t1
            gamct = -gamc[ind]*t1
            egamt = np.exp(gamt)
            ec = egamt*c2[ind] - np.exp(gamct)*c1[ind]
            #NxQ terms
            t_lq = t1/lq

            # Upsilon Calculations
            # Using wofz
            wnu = wofz(1j*nu)
            lwnu = np.log(wnu)
            t2_lq2 = -t_lq*t_lq
            upm = wnu[ind] - np.exp(t2_lq2 + gamt + np.log(wofz(1j*(t_lq + nu[ind]))))
            upm[t1[:, 0] == 0, :] = 0.

            nu2 = nu*nu
            z1 = nu[ind] - t_lq
            indv1 = np.where(z1.real >= 0.)
            indv2 = np.where(z1.real < 0.)
            upv = -np.exp(lwnu[ind] + gamt)
            if indv1[0].shape > 0:
                upv[indv1] += np.exp(t2_lq2[indv1] + np.log(wofz(1j*z1[indv1])))
            if indv2[0].shape > 0:
                upv[indv2] += np.exp(nu2[ind[indv2[0]], indv2[1]] + gamt[indv2[0], 0] + np.log(2.))\
                             - np.exp(t2_lq2[indv2] + np.log(wofz(-1j*z1[indv2])))
            upv[t1[:, 0] == 0, :] = 0.

            #Covariance calculation
            kdiag[ind3t] = np.real(K01[ind]*upm)
            kdiag[ind3t] += np.real((c0[ind]*ec)*upv)

        #(2) When w_d is complex
        if np.any(wbool):
            t1 = t[ind2t]
            ind = index[ind2t]
            #Index transformation
            d = np.asarray(np.where(wbool)[0])
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            S2lq = S2[d]*(lq*.25)
            c0 = S2lq*np.sqrt(np.pi)
            w = .5*np.sqrt(C2[d] - 4.*B[d])
            alphad = alpha[d]
            gam = alphad - w
            gamc = alphad + w
            w2 = -w*w
            c1 = .5/(alphad*w2)
            c21 = .5/(gam*w2)
            c22 = .5/(gamc*w2)
            c = c1 - c21
            c2 = c1 - c22
            #DxQ terms
            K011 = c0*c
            K012 = c0*c2
            nu = lq*(.5*gam)
            nuc = lq*(.5*gamc)
            #Nx1 terms
            gamt = -gam[ind]*t1
            gamct = -gamc[ind]*t1
            egamt = np.exp(gamt)
            egamct = np.exp(gamct)
            ec = egamt*c21[ind] - egamct*c1[ind]
            ec2 = egamct*c22[ind] - egamt*c1[ind]
            #NxQ terms
            t_lq = t1/lq

            #Upsilon Calculations using wofz
            t2_lq2 = -t_lq*t_lq #Required when using wofz
            wnu = wofz(1j*nu).real
            lwnu = np.log(wnu)
            upm = wnu[ind] - np.exp(t2_lq2 + gamt + np.log(wofz(1j*(t_lq + nu[ind])).real))
            upm[t1[:, 0] == 0., :] = 0.

            nu2 = nu*nu
            z1 = nu[ind] - t_lq
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            upv = -np.exp(lwnu[ind] + gamt)
            if indv1[0].shape > 0:
                upv[indv1] += np.exp(t2_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                upv[indv2] += np.exp(nu2[ind[indv2[0]], indv2[1]] + gamt[indv2[0], 0] + np.log(2.))\
                              - np.exp(t2_lq2[indv2] + np.log(wofz(-1j*z1[indv2]).real))
            upv[t1[:, 0] == 0, :] = 0.

            wnuc = wofz(1j*nuc).real
            lwnuc = np.log(wnuc)

            upmc = wnuc[ind] - np.exp(t2_lq2 + gamct + np.log(wofz(1j*(t_lq + nuc[ind])).real))
            upmc[t1[:, 0] == 0., :] = 0.

            nuc2 = nuc*nuc
            z1 = nuc[ind] - t_lq
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            upvc = - np.exp(lwnuc[ind] + gamct)
            if indv1[0].shape > 0:
                upvc[indv1] += np.exp(t2_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                upvc[indv2] += np.exp(nuc2[ind[indv2[0]], indv2[1]] + gamct[indv2[0], 0] + np.log(2.))\
                               - np.exp(t2_lq2[indv2] + np.log(wofz(-1j*z1[indv2]).real))
            upvc[t1[:, 0] == 0, :] = 0.

            #Covariance calculation
            kdiag[ind2t] = K011[ind]*upm + K012[ind]*upmc + (c0[ind]*ec)*upv + (c0[ind]*ec2)*upvc
        return kdiag

    def update_gradients_full(self, dL_dK, X, X2 = None):
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        self.B.gradient = np.zeros(self.B.shape)
        self.C.gradient = np.zeros(self.C.shape)
        self.W.gradient = np.zeros(self.W.shape)
        self.lengthscale.gradient = np.zeros(self.lengthscale.shape)
        index = np.int_(X[:, 1])
        index = index.reshape(index.size,)
        X_flag = index[0] >= self.output_dim
        if X2 is None:
            if X_flag: #Kuu or Kmm
                index -= self.output_dim
                tmp = dL_dK*self._gkuu_lq(X, index)
                for q in np.unique(index):
                    ind = np.where(index == q)
                    self.lengthscale.gradient[q] = tmp[np.ix_(ind[0], ind[0])].sum()
            else:
                raise NotImplementedError
        else: #Kfu or Knm
            #index2 = np.asarray(X2, dtype=np.int)
            #index2 = index2.reshape(index2.size,)
            if hasattr(X2, 'values'):
                X2 = X2.values
            index2 = np.int_(X2[:, 1])
            index2 = index2.reshape(index2.size,)
            X2_flag = index2[0] >= self.output_dim
            if not X_flag and X2_flag:
                index2 -= self.output_dim
            else:
                dL_dK = dL_dK.T #so we obtaing dL_Kfu
                indtemp = index - self.output_dim
                Xtemp = X
                X = X2
                X2 = Xtemp
                index = index2
                index2 = indtemp
            glq, gSdq, gB, gC = self._gkfu(X, index, X2, index2)
            tmp = dL_dK*glq
            for q in np.unique(index2):
                ind = np.where(index2 == q)
                self.lengthscale.gradient[q] = tmp[:, ind].sum()
            tmpB = dL_dK*gB
            tmpC = dL_dK*gC
            tmp = dL_dK*gSdq
            for d in np.unique(index):
                ind = np.where(index == d)
                self.B.gradient[d] = tmpB[ind, :].sum()
                self.C.gradient[d] = tmpC[ind, :].sum()
                for q in np.unique(index2):
                    ind2 = np.where(index2 == q)
                    self.W.gradient[d, q] = tmp[np.ix_(ind[0], ind2[0])].sum()

    def update_gradients_diag(self, dL_dKdiag, X):
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        self.B.gradient = np.zeros(self.B.shape)
        self.C.gradient = np.zeros(self.C.shape)
        self.W.gradient = np.zeros(self.W.shape)
        self.lengthscale.gradient = np.zeros(self.lengthscale.shape)
        index = np.int_(X[:, 1])
        index = index.reshape(index.size,)
        
        glq, gS, gB, gC = self._gkdiag(X, index)
        if dL_dKdiag.size == X.shape[0]:
            dL_dKdiag = np.reshape(dL_dKdiag, (index.size, 1))
        tmp = dL_dKdiag*glq
        self.lengthscale.gradient = tmp.sum(0)
        tmpB = dL_dKdiag*gB
        tmpC = dL_dKdiag*gC
        tmp = dL_dKdiag*gS
        for d in np.unique(index):
            ind = np.where(index == d)
            self.B.gradient[d] = tmpB[ind, :].sum()
            self.C.gradient[d] = tmpC[ind, :].sum()
            self.W.gradient[d, :] = tmp[ind].sum(0)

    def gradients_X(self, dL_dK, X, X2=None):
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        index = np.int_(X[:, 1])
        index = index.reshape(index.size,)
        X_flag = index[0] >= self.output_dim
        #If input_dim == 1, use this
        #gX = np.zeros((X.shape[0], 1))
        #Cheat to allow gradient for input_dim==2
        gX = np.zeros(X.shape)
        if X2 is None: #Kuu or Kmm
            if X_flag:
                index -= self.output_dim
                gX[:, 0] = 2.*(dL_dK*self._gkuu_X(X, index)).sum(0)
                return gX
            else:
                raise NotImplementedError
        else: #Kuf or Kmn
            #index2 = np.asarray(X2, dtype=np.int)
            #index2 = index2.reshape(index2.size,)
            if hasattr(X2, 'values'):
                X2 = X2.values
            index2 = np.int_(X2[:, 1])
            index2 = index2.reshape(index2.size,)
            X2_flag = index2[0] >= self.output_dim
            if X_flag and not X2_flag: #gradient of Kuf(Z, X) wrt Z
                index -= self.output_dim
                gX[:, 0] = (dL_dK*self._gkfu_z(X2, index2, X, index).T).sum(1)
                return gX
            else:
                raise NotImplementedError

    #---------------------------------------#
    #             Helper functions          #
    #---------------------------------------#

    #Evaluation of squared exponential for LFM
    def _Kuu(self, X, index):
        index = index.reshape(index.size,)
        t = X[:, 0].reshape(X.shape[0],)
        lq = self.lengthscale.values.reshape(self.rank,)
        lq2 = lq*lq
        #Covariance matrix initialization
        kuu = np.zeros((t.size, t.size))
        #Assign 1. to diagonal terms
        kuu[np.diag_indices(t.size)] = 1.
        #Upper triangular indices
        indtri1, indtri2 = np.triu_indices(t.size, 1)
        #Block Diagonal indices among Upper Triangular indices
        ind = np.where(index[indtri1] == index[indtri2])
        indr = indtri1[ind]
        indc = indtri2[ind]
        r = t[indr] - t[indc]
        r2 = r*r
        #Calculation of  covariance function
        kuu[indr, indc] = np.exp(-r2/lq2[index[indr]])
        #Completation of lower triangular part
        kuu[indc, indr] = kuu[indr, indc]
        return kuu

    def _Kusu(self, X, index, X2, index2):
        index = index.reshape(index.size,)
        index2 = index2.reshape(index2.size,)
        t = X[:, 0].reshape(X.shape[0],1)
        t2 = X2[:, 0].reshape(1,X2.shape[0])
        lq = self.lengthscale.values.reshape(self.rank,)
        #Covariance matrix initialization
        kuu = np.zeros((t.size, t2.size))
        for q in range(self.rank):
            ind1 = index == q
            ind2 = index2 == q
            r = t[ind1]/lq[q] - t2[0,ind2]/lq[q]
            r2 = r*r
            #Calculation of  covariance function
            kuu[np.ix_(ind1, ind2)] = np.exp(-r2)
        return kuu

    #Evaluation of cross-covariance function
    def _Kfu(self, X, index, X2, index2):
        #terms that move along t
        t = X[:, 0].reshape(X.shape[0], 1)
        d = np.unique(index) #Output Indexes
        B = self.B.values[d]
        C = self.C.values[d]
        S = self.W.values[d, :]
        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]
        #Check where wd becomes complex
        wbool = C*C >= 4.*B
        #Output related variables must be column-wise
        C = C.reshape(C.size, 1)
        B = B.reshape(B.size, 1)
        C2 = C*C
        #Input related variables must be row-wise
        z = X2[:, 0].reshape(1, X2.shape[0])
        lq = self.lengthscale.values.reshape((1, self.rank))
        #print np.max(z), np.max(z/lq[0, index2])
        alpha = .5*C

        wbool2 = wbool[index]
        ind2t = np.where(wbool2)
        ind3t = np.where(np.logical_not(wbool2))

        kfu = np.empty((t.size, z.size))

        indD = np.arange(B.size)
        #(1) when wd is real
        if np.any(np.logical_not(wbool)):
            #Indexes of index and t related to (2)
            t1 = t[ind3t]
            ind = index[ind3t]
            #Index transformation
            d = np.asarray(np.where(np.logical_not(wbool))[0])
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            w = .5*np.sqrt(4.*B[d] - C2[d])
            alphad = alpha[d]
            gam = alphad - 1j*w

            #DxQ terms
            Slq = (S[d]/w)*(.5*lq)
            c0 = Slq*np.sqrt(np.pi)
            nu = gam*(.5*lq)
            #1xM terms
            z_lq = z/lq[0, index2]
            #NxQ terms
            t_lq = t1/lq
            #NxM terms
            zt_lq = z_lq - t_lq[:, index2]

            # Upsilon Calculations
            #Using wofz
            tz = t1-z
            fullind = np.ix_(ind, index2)
            zt_lq2 = -zt_lq*zt_lq
            z_lq2 = -z_lq*z_lq
            gamt = -gam[ind]*t1

            upsi = - np.exp(z_lq2 + gamt + np.log(wofz(1j*(z_lq + nu[fullind]))))
            z1 = zt_lq + nu[fullind]
            indv1 = np.where(z1.real >= 0.)
            indv2 = np.where(z1.real < 0.)
            if indv1[0].shape > 0:
                upsi[indv1] += np.exp(zt_lq2[indv1] + np.log(wofz(1j*z1[indv1])))
            if indv2[0].shape > 0:
                nua2 = nu[ind[indv2[0]], index2[indv2[1]]]**2
                upsi[indv2] += np.exp(nua2 - gam[ind[indv2[0]], 0]*tz[indv2] + np.log(2.))\
                               - np.exp(zt_lq2[indv2] + np.log(wofz(-1j*z1[indv2])))
            upsi[t1[:, 0] == 0., :] = 0.

            #Covariance calculation
            kfu[ind3t] = c0[fullind]*upsi.imag

        #(2) when wd is complex
        if np.any(wbool):
            #Indexes of index and t related to (2)
            t1 = t[ind2t]
            ind = index[ind2t]
            #Index transformation
            d = np.asarray(np.where(wbool)[0])
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            w = .5*np.sqrt(C2[d] - 4.*B[d])
            alphad = alpha[d]
            gam = alphad - w
            gamc = alphad + w
            #DxQ terms
            Slq = S[d]*(lq*.25)
            c0 = -Slq*(np.sqrt(np.pi)/w)
            nu = gam*(lq*.5)
            nuc = gamc*(lq*.5)
            #1xM terms
            z_lq = z/lq[0, index2]
            #NxQ terms
            t_lq = t1/lq[0, index2]
            #NxM terms
            zt_lq = z_lq - t_lq

            # Upsilon Calculations
            tz = t1-z
            z_lq2 = -z_lq*z_lq
            zt_lq2 = -zt_lq*zt_lq
            gamt = -gam[ind]*t1
            gamct = -gamc[ind]*t1
            fullind = np.ix_(ind, index2)
            upsi = np.exp(z_lq2 + gamt + np.log(wofz(1j*(z_lq + nu[fullind])).real))\
                   - np.exp(z_lq2 + gamct + np.log(wofz(1j*(z_lq + nuc[fullind])).real))

            z1 = zt_lq + nu[fullind]
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            if indv1[0].shape > 0:
                upsi[indv1] -= np.exp(zt_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                nua2 = nu[ind[indv2[0]], index2[indv2[1]]]**2
                upsi[indv2] -= np.exp(nua2 - gam[ind[indv2[0]], 0]*tz[indv2] + np.log(2.))\
                               - np.exp(zt_lq2[indv2] + np.log(wofz(-1j*z1[indv2]).real))
            z1 = zt_lq + nuc[fullind]
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            if indv1[0].shape > 0:
                upsi[indv1] += np.exp(zt_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                nuac2 = nuc[ind[indv2[0]], index2[indv2[1]]]**2
                upsi[indv2] += np.exp(nuac2 - gamc[ind[indv2[0]], 0]*tz[indv2] + np.log(2.))\
                               - np.exp(zt_lq2[indv2] + np.log(wofz(-1j*z1[indv2]).real))
            upsi[t1[:, 0] == 0., :] = 0.

            kfu[ind2t] = c0[np.ix_(ind, index2)]*upsi
        return kfu

    #Gradient of Kuu wrt lengthscale
    def _gkuu_lq(self, X, index):
        t = X[:, 0].reshape(X.shape[0],)
        index = index.reshape(X.shape[0],)
        lq = self.lengthscale.values.reshape(self.rank,)
        lq2 = lq*lq
        #Covariance matrix initialization
        glq = np.zeros((t.size, t.size))
        #Upper triangular indices
        indtri1, indtri2 = np.triu_indices(t.size, 1)
        #Block Diagonal indices among Upper Triangular indices
        ind = np.where(index[indtri1] == index[indtri2])
        indr = indtri1[ind]
        indc = indtri2[ind]
        r = t[indr] - t[indc]
        r2 = r*r
        r2_lq2 = r2/lq2[index[indr]]
        #Calculation of  covariance function
        er2_lq2 = np.exp(-r2_lq2)
        #Gradient wrt lq
        c = 2.*r2_lq2/lq[index[indr]]
        glq[indr, indc] = er2_lq2*c
        #Complete the lower triangular
        glq[indc, indr] = glq[indr, indc]
        return glq

    #Be careful this derivative should be transpose it
    def _gkuu_X(self, X, index): #Diagonal terms are always zero
        t = X[:, 0].reshape(X.shape[0],)
        index = index.reshape(index.size,)
        lq = self.lengthscale.values.reshape(self.rank,)
        lq2 = lq*lq
        #Covariance matrix initialization
        gt = np.zeros((t.size, t.size))
        #Upper triangular indices
        indtri1, indtri2 = np.triu_indices(t.size, 1) #Offset of 1 from the diagonal
        #Block Diagonal indices among Upper Triangular indices
        ind = np.where(index[indtri1] == index[indtri2])
        indr = indtri1[ind]
        indc = indtri2[ind]
        r = t[indr] - t[indc]
        r2 = r*r
        r2_lq2 = r2/(-lq2[index[indr]])
        #Calculation of  covariance function
        er2_lq2 = np.exp(r2_lq2)
        #Gradient wrt t
        c = 2.*r/lq2[index[indr]]
        gt[indr, indc] = er2_lq2*c
        #Complete the lower triangular
        gt[indc, indr] = -gt[indr, indc]
        return gt

    #Gradients for Diagonal Kff
    def _gkdiag(self, X, index):
        index = index.reshape(index.size,)
        #terms that move along t
        d = np.unique(index)
        B = self.B[d].values
        C = self.C[d].values
        S = self.W[d, :].values
        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]
        #Check where wd becomes complex
        wbool = C*C >= 4.*B
        #Output related variables must be column-wise
        t = X[:, 0].reshape(X.shape[0], 1)
        B = B.reshape(B.size, 1)
        C = C.reshape(C.size, 1)
        alpha = .5*C
        C2 = C*C
        S2 = S*S

        wbool2 = wbool[index]
        ind2t = np.where(wbool2)
        ind3t = np.where(np.logical_not(wbool2))

        #Input related variables must be row-wise
        lq = self.lengthscale.values.reshape(1, self.rank)
        lq2 = lq*lq

        gB = np.empty((t.size, lq.size))
        gC = np.empty((t.size, lq.size))
        glq = np.empty((t.size, lq.size))
        gS = np.empty((t.size, lq.size))

        indD = np.arange(B.size)
        #(1) When wd is real
        if np.any(np.logical_not(wbool)):
            #Indexes of index and t related to (1)
            t1 = t[ind3t]
            ind = index[ind3t]
            #Index transformation
            d = np.asarray(np.where(np.logical_not(wbool))[0])
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            S2lq = S2[d]*(.5*lq)
            c0 = S2lq*np.sqrt(np.pi)

            w = .5*np.sqrt(4.*B[d] - C2[d])
            alphad = alpha[d]
            alpha2 = alphad*alphad
            w2 = w*w
            gam = alphad + 1j*w
            gam2 = gam*gam
            gamc = alphad - 1j*w
            c1 = 0.5/alphad
            c2 = 0.5/gam
            c = c1 - c2

            #DxQ terms
            c0 = c0/w2
            nu = (.5*lq)*gam
            #Nx1 terms
            gamt = -gam[ind]*t1
            gamct = -gamc[ind]*t1
            egamt = np.exp(gamt)
            egamct = np.exp(gamct)
            ec = egamt*c2[ind] - egamct*c1[ind]

            #NxQ terms
            t_lq = t1/lq
            t2_lq2 = -t_lq*t_lq
            t_lq2 = t_lq/lq

            et2_lq2 = np.exp(t2_lq2)
            etlq2gamt = np.exp(t2_lq2 + gamt)

            ##Upsilon calculations
            #Using wofz
            wnu = wofz(1j*nu)
            lwnu = np.log(wnu)
            t2_lq2 = -t_lq*t_lq
            upm = wnu[ind] - np.exp(t2_lq2 + gamt + np.log(wofz(1j*(t_lq + nu[ind]))))
            upm[t1[:, 0] == 0, :] = 0.

            nu2 = nu*nu
            z1 = nu[ind] - t_lq
            indv1 = np.where(z1.real >= 0.)
            indv2 = np.where(z1.real < 0.)
            upv = -np.exp(lwnu[ind] + gamt)
            if indv1[0].shape > 0:
                upv[indv1] += np.exp(t2_lq2[indv1] + np.log(wofz(1j*z1[indv1])))
            if indv2[0].shape > 0:
                upv[indv2] += np.exp(nu2[ind[indv2[0]], indv2[1]] + gamt[indv2[0], 0] + np.log(2.))\
                             - np.exp(t2_lq2[indv2] + np.log(wofz(-1j*z1[indv2])))
            upv[t1[:, 0] == 0, :] = 0.

            #Gradient wrt S
            Slq = S[d]*lq #For grad wrt S
            c0_S = Slq*np.sqrt(np.pi)/w2
            K01 = c0_S*c

            gS[ind3t] = np.real(K01[ind]*upm) + np.real((c0_S[ind]*ec)*upv)

            #For B and C
            upmd = etlq2gamt - 1.
            upvd = egamt - et2_lq2

            # gradient wrt B
            dw_dB = 0.5/w
            dgam_dB = 1j*dw_dB

            Ba1 = c0*(0.5*dgam_dB/gam2 + (0.5*lq2*gam*dgam_dB - 2.*dw_dB/w)*c)
            Ba2_1 = c0*(dgam_dB*(0.5/gam2 - 0.25*lq2) + dw_dB/(w*gam))
            Ba2_2 = c0*dgam_dB/gam
            Ba3 = c0*(-0.25*lq2*gam*dgam_dB/alphad + dw_dB/(w*alphad))
            Ba4_1 = (S2lq*lq)*dgam_dB/w2
            Ba4 = Ba4_1*c

            gB[ind3t] = np.real(Ba1[ind]*upm) - np.real(((Ba2_1[ind] + Ba2_2[ind]*t1)*egamt - Ba3[ind]*egamct)*upv)\
                + np.real(Ba4[ind]*upmd) + np.real((Ba4_1[ind]*ec)*upvd)

            # gradient wrt C
            dw_dC = - alphad*dw_dB
            dgam_dC = 0.5 + 1j*dw_dC

            Ca1 = c0*(-0.25/alpha2 + 0.5*dgam_dC/gam2 + (0.5*lq2*gam*dgam_dC - 2.*dw_dC/w)*c)
            Ca2_1 = c0*(dgam_dC*(0.5/gam2 - 0.25*lq2) + dw_dC/(w*gam))
            Ca2_2 = c0*dgam_dC/gam
            Ca3_1 = c0*(0.25/alpha2 - 0.25*lq2*gam*dgam_dC/alphad + dw_dC/(w*alphad))
            Ca3_2 = 0.5*c0/alphad
            Ca4_1 = (S2lq*lq)*dgam_dC/w2
            Ca4 = Ca4_1*c

            gC[ind3t] = np.real(Ca1[ind]*upm) - np.real(((Ca2_1[ind] + Ca2_2[ind]*t1)*egamt - (Ca3_1[ind] + Ca3_2[ind]*t1)*egamct)*upv)\
                + np.real(Ca4[ind]*upmd) + np.real((Ca4_1[ind]*ec)*upvd)

            #Gradient wrt lengthscale
            #DxQ terms
            la = (1./lq + nu*gam)*c0
            la1 = la*c

            c0l = (S2[d]/w2)*lq
            la3 = c0l*c
            gam_2 = .5*gam
            glq[ind3t] = (la1[ind]*upm).real + ((la[ind]*ec)*upv).real\
                + (la3[ind]*(-gam_2[ind] + etlq2gamt*(-t_lq2 + gam_2[ind]))).real\
                + ((c0l[ind]*ec)*(-et2_lq2*(t_lq2 + gam_2[ind]) + egamt*gam_2[ind])).real

        #(2) When w_d is complex
        if np.any(wbool):
            t1 = t[ind2t]
            ind = index[ind2t]
            #Index transformation
            d = np.asarray(np.where(wbool)[0])
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            S2lq = S2[d]*(.25*lq)
            c0 = S2lq*np.sqrt(np.pi)
            w = .5*np.sqrt(C2[d]-4.*B[d])
            w2 = -w*w
            alphad = alpha[d]
            alpha2 = alphad*alphad
            gam = alphad - w
            gamc = alphad + w
            gam2 = gam*gam
            gamc2 = gamc*gamc
            c1 = .5/alphad
            c21 = .5/gam
            c22 = .5/gamc
            c = c1 - c21
            c2 = c1 - c22
            #DxQ terms
            c0 = c0/w2
            nu = .5*lq*gam
            nuc = .5*lq*gamc

            #Nx1 terms
            gamt = -gam[ind]*t1
            gamct = -gamc[ind]*t1
            egamt = np.exp(gamt)
            egamct = np.exp(gamct)
            ec = egamt*c21[ind] - egamct*c1[ind]
            ec2 = egamct*c22[ind] - egamt*c1[ind]
            #NxQ terms
            t_lq = t1/lq
            t2_lq2 = -t_lq*t_lq

            et2_lq2 = np.exp(t2_lq2)
            etlq2gamct = np.exp(t2_lq2 + gamct)
            etlq2gamt = np.exp(t2_lq2 + gamt)

            #Upsilon Calculations using wofz
            t2_lq2 = -t_lq*t_lq #Required when using wofz
            wnu = np.real(wofz(1j*nu))
            lwnu = np.log(wnu)

            upm = wnu[ind] - np.exp(t2_lq2 + gamt + np.log(wofz(1j*(t_lq + nu[ind])).real))
            upm[t1[:, 0] == 0., :] = 0.

            nu2 = nu*nu
            z1 = nu[ind] - t_lq
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            upv = -np.exp(lwnu[ind] + gamt)
            if indv1[0].shape > 0:
                upv[indv1] += np.exp(t2_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                upv[indv2] += np.exp(nu2[ind[indv2[0]], indv2[1]] + gamt[indv2[0], 0] + np.log(2.)) - np.exp(t2_lq2[indv2]\
                    + np.log(wofz(-1j*z1[indv2]).real))
            upv[t1[:, 0] == 0, :] = 0.

            wnuc = wofz(1j*nuc).real
            upmc = wnuc[ind] - np.exp(t2_lq2 + gamct + np.log(wofz(1j*(t_lq + nuc[ind])).real))
            upmc[t1[:, 0] == 0., :] = 0.

            lwnuc = np.log(wnuc)
            nuc2 = nuc*nuc
            z1 = nuc[ind] - t_lq
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            upvc = -np.exp(lwnuc[ind] + gamct)
            if indv1[0].shape > 0:
                upvc[indv1] += np.exp(t2_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                upvc[indv2] += np.exp(nuc2[ind[indv2[0]], indv2[1]] + gamct[indv2[0], 0] + np.log(2.)) - np.exp(t2_lq2[indv2]\
                    + np.log(wofz(-1j*z1[indv2]).real))
            upvc[t1[:, 0] == 0, :] = 0.

            #Gradient wrt S
            #NxQ terms
            c0_S = (S[d]/w2)*(lq*(np.sqrt(np.pi)*.5))

            K011 = c0_S*c
            K012 = c0_S*c2

            gS[ind2t] = K011[ind]*upm + K012[ind]*upmc + (c0_S[ind]*ec)*upv + (c0_S[ind]*ec2)*upvc

            #Is required to cache this, C gradient also required them
            upmd = -1. + etlq2gamt
            upvd = -et2_lq2 + egamt
            upmdc = -1. + etlq2gamct
            upvdc = -et2_lq2 + egamct

            # Gradient wrt B
            dgam_dB = 0.5/w
            dgamc_dB = -dgam_dB

            Ba1 = c0*(0.5*dgam_dB/gam2 + (0.5*lq2*gam*dgam_dB - 1./w2)*c)
            Ba3 = c0*(-0.25*lq2*gam*dgam_dB/alphad + 0.5/(w2*alphad))
            Ba4_1 = (S2lq*lq)*dgam_dB/w2
            Ba4 = Ba4_1*c
            Ba2_1 = c0*(dgam_dB*(0.5/gam2 - 0.25*lq2) + 0.5/(w2*gam))
            Ba2_2 = c0*dgam_dB/gam

            Ba1c = c0*(0.5*dgamc_dB/gamc2 + (0.5*lq2*gamc*dgamc_dB - 1./w2)*c2)
            Ba3c = c0*(-0.25*lq2*gamc*dgamc_dB/alphad + 0.5/(w2*alphad))
            Ba4_1c = (S2lq*lq)*dgamc_dB/w2
            Ba4c = Ba4_1c*c2
            Ba2_1c = c0*(dgamc_dB*(0.5/gamc2 - 0.25*lq2) + 0.5/(w2*gamc))
            Ba2_2c = c0*dgamc_dB/gamc

            gB[ind2t] = Ba1[ind]*upm - ((Ba2_1[ind] + Ba2_2[ind]*t1)*egamt - Ba3[ind]*egamct)*upv\
                + Ba4[ind]*upmd + (Ba4_1[ind]*ec)*upvd\
                + Ba1c[ind]*upmc - ((Ba2_1c[ind] + Ba2_2c[ind]*t1)*egamct - Ba3c[ind]*egamt)*upvc\
                + Ba4c[ind]*upmdc + (Ba4_1c[ind]*ec2)*upvdc

            ##Gradient wrt C
            dw_dC = 0.5*alphad/w
            dgam_dC = 0.5 - dw_dC
            dgamc_dC = 0.5 + dw_dC
            S2lq2 = S2lq*lq

            Ca1 = c0*(-0.25/alpha2 + 0.5*dgam_dC/gam2 + (0.5*lq2*gam*dgam_dC + alphad/w2)*c)
            Ca2_1 = c0*(dgam_dC*(0.5/gam2 - 0.25*lq2) - 0.5*alphad/(w2*gam))
            Ca2_2 = c0*dgam_dC/gam
            Ca3_1 = c0*(0.25/alpha2 - 0.25*lq2*gam*dgam_dC/alphad - 0.5/w2)
            Ca3_2 = 0.5*c0/alphad
            Ca4_1 = S2lq2*(dgam_dC/w2)
            Ca4 = Ca4_1*c

            Ca1c = c0*(-0.25/alpha2 + 0.5*dgamc_dC/gamc2 + (0.5*lq2*gamc*dgamc_dC + alphad/w2)*c2)
            Ca2_1c = c0*(dgamc_dC*(0.5/gamc2 - 0.25*lq2) - 0.5*alphad/(w2*gamc))
            Ca2_2c = c0*dgamc_dC/gamc
            Ca3_1c = c0*(0.25/alpha2 - 0.25*lq2*gamc*dgamc_dC/alphad - 0.5/w2)
            Ca3_2c = 0.5*c0/alphad
            Ca4_1c = S2lq2*(dgamc_dC/w2)
            Ca4c = Ca4_1c*c2

            gC[ind2t] = Ca1[ind]*upm - ((Ca2_1[ind] + Ca2_2[ind]*t1)*egamt - (Ca3_1[ind] + Ca3_2[ind]*t1)*egamct)*upv\
                + Ca4[ind]*upmd + (Ca4_1[ind]*ec)*upvd\
                + Ca1c[ind]*upmc - ((Ca2_1c[ind] + Ca2_2c[ind]*t1)*egamct - (Ca3_1c[ind] + Ca3_2c[ind]*t1)*egamt)*upvc\
                + Ca4c[ind]*upmdc + (Ca4_1c[ind]*ec2)*upvdc

            #Gradient wrt lengthscale
            #DxQ terms
            la = (1./lq + nu*gam)*c0
            lac = (1./lq + nuc*gamc)*c0
            la1 = la*c
            la1c = lac*c2
            t_lq2 = t_lq/lq
            c0l = (S2[d]/w2)*(.5*lq)
            la3 = c0l*c
            la3c = c0l*c2
            gam_2 = .5*gam
            gamc_2 = .5*gamc
            glq[ind2t] = la1c[ind]*upmc + (lac[ind]*ec2)*upvc\
                + la3c[ind]*(-gamc_2[ind] + etlq2gamct*(-t_lq2 + gamc_2[ind]))\
                + (c0l[ind]*ec2)*(-et2_lq2*(t_lq2 + gamc_2[ind]) + egamct*gamc_2[ind])\
                + la1[ind]*upm + (la[ind]*ec)*upv\
                + la3[ind]*(-gam_2[ind] + etlq2gamt*(-t_lq2 + gam_2[ind]))\
                + (c0l[ind]*ec)*(-et2_lq2*(t_lq2 + gam_2[ind]) + egamt*gam_2[ind])

        return glq, gS, gB, gC

    def _gkfu(self, X, index, Z, index2):
        index = index.reshape(index.size,)
        #TODO: reduce memory usage
        #terms that move along t
        d = np.unique(index)
        B = self.B[d].values
        C = self.C[d].values
        S = self.W[d, :].values
        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]
        #Check where wd becomes complex
        wbool = C*C >= 4.*B
        #t column
        t = X[:, 0].reshape(X.shape[0], 1)
        C = C.reshape(C.size, 1)
        B = B.reshape(B.size, 1)
        C2 = C*C
        #z row
        z = Z[:, 0].reshape(1, Z.shape[0])
        index2 = index2.reshape(index2.size,)
        lq = self.lengthscale.values.reshape((1, self.rank))
        lq2 = lq*lq

        alpha = .5*C

        wbool2 = wbool[index]
        ind2t = np.where(wbool2)
        ind3t = np.where(np.logical_not(wbool2))
        #kfu = np.empty((t.size, z.size))
        glq = np.empty((t.size, z.size))
        gSdq = np.empty((t.size, z.size))
        gB = np.empty((t.size, z.size))
        gC = np.empty((t.size, z.size))

        indD = np.arange(B.size)
        #(1) when wd is real
        if np.any(np.logical_not(wbool)):
            #Indexes of index and t related to (2)
            t1 = t[ind3t]
            ind = index[ind3t]
            #Index transformation
            d = np.asarray(np.where(np.logical_not(wbool))[0])
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            w = .5*np.sqrt(4.*B[d] - C2[d])
            alphad = alpha[d]
            gam = alphad - 1j*w
            gam_2 = .5*gam
            S_w = S[d]/w
            S_wpi = S_w*(.5*np.sqrt(np.pi))
            #DxQ terms
            c0 = S_wpi*lq #lq*Sdq*sqrt(pi)/(2w)
            nu = gam*lq
            nu2 = 1.+.5*(nu*nu)
            nu *= .5

            #1xM terms
            z_lq = z/lq[0, index2]
            z_lq2 = -z_lq*z_lq
            #NxQ terms
            t_lq = t1/lq
            #DxM terms
            gamt = -gam[ind]*t1
            #NxM terms
            zt_lq = z_lq - t_lq[:, index2]
            zt_lq2 = -zt_lq*zt_lq
            ezt_lq2 = -np.exp(zt_lq2)
            ezgamt = np.exp(z_lq2 + gamt)

            # Upsilon calculations
            fullind = np.ix_(ind, index2)
            upsi = - np.exp(z_lq2 + gamt + np.log(wofz(1j*(z_lq + nu[fullind]))))
            tz = t1-z
            z1 = zt_lq + nu[fullind]
            indv1 = np.where(z1.real >= 0.)
            indv2 = np.where(z1.real < 0.)
            if indv1[0].shape > 0:
                upsi[indv1] += np.exp(zt_lq2[indv1] + np.log(wofz(1j*z1[indv1])))
            if indv2[0].shape > 0:
                nua2 = nu[ind[indv2[0]], index2[indv2[1]]]**2
                upsi[indv2] += np.exp(nua2 - gam[ind[indv2[0]], 0]*tz[indv2] + np.log(2.))\
                               - np.exp(zt_lq2[indv2] + np.log(wofz(-1j*z1[indv2])))
            upsi[t1[:, 0] == 0., :] = 0.

            #Gradient wrt S
            #DxQ term
            Sa1 = lq*(.5*np.sqrt(np.pi))/w

            gSdq[ind3t] = Sa1[np.ix_(ind, index2)]*upsi.imag

            #Gradient wrt lq
            la1 = S_wpi*nu2
            la2 = S_w*lq
            uplq = ezt_lq2*(gam_2[ind])
            uplq += ezgamt*(-z_lq/lq[0, index2] + gam_2[ind])

            glq[ind3t] = (la1[np.ix_(ind, index2)]*upsi).imag
            glq[ind3t] += la2[np.ix_(ind, index2)]*uplq.imag

            #Gradient wrt B
            #Dx1 terms
            dw_dB = .5/w
            dgam_dB = -1j*dw_dB
            #DxQ terms
            Ba1 = -c0*dw_dB/w #DXQ
            Ba2 = c0*dgam_dB #DxQ
            Ba3 = lq2*gam_2 #DxQ
            Ba4 = (dgam_dB*S_w)*(.5*lq2) #DxQ

            gB[ind3t] = ((Ba1[np.ix_(ind, index2)] + Ba2[np.ix_(ind, index2)]*(Ba3[np.ix_(ind, index2)] - (t1-z)))*upsi).imag\
                + (Ba4[np.ix_(ind, index2)]*(ezt_lq2 + ezgamt)).imag

            #Gradient wrt C (it uses some calculations performed in B)
            #Dx1 terms
            dw_dC = -.5*alphad/w
            dgam_dC = 0.5 - 1j*dw_dC
            #DxQ terms
            Ca1 = -c0*dw_dC/w #DXQ
            Ca2 = c0*dgam_dC #DxQ
            Ca4 = (dgam_dC*S_w)*(.5*lq2) #DxQ

            gC[ind3t] = ((Ca1[np.ix_(ind, index2)] + Ca2[np.ix_(ind, index2)]*(Ba3[np.ix_(ind, index2)] - (t1-z)))*upsi).imag\
                + (Ca4[np.ix_(ind, index2)]*(ezt_lq2 + ezgamt)).imag

        #(2) when wd is complex
        if np.any(wbool):
            #Indexes of index and t related to (2)
            t1 = t[ind2t]
            ind = index[ind2t]
            #Index transformation
            d = np.asarray(np.where(wbool)[0])
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            w = .5*np.sqrt(C2[d] - 4.*B[d])
            w2 = w*w
            alphad = alpha[d]
            gam = alphad - w
            gamc = alphad + w
            #DxQ terms
            S_w= -S[d]/w #minus is given by j*j
            S_wpi = S_w*(.25*np.sqrt(np.pi))

            c0 = S_wpi*lq
            gam_2 = .5*gam
            gamc_2 = .5*gamc
            nu = gam*lq
            nuc = gamc*lq
            nu2 = 1.+.5*(nu*nu)
            nuc2 = 1.+.5*(nuc*nuc)
            nu *= .5
            nuc *= .5
            #1xM terms
            z_lq = z/lq[0, index2]
            z_lq2 = -z_lq*z_lq
            #Nx1
            gamt = -gam[ind]*t1
            gamct = -gamc[ind]*t1
            #NxQ terms
            t_lq = t1/lq[0, index2]
            #NxM terms
            zt_lq = z_lq - t_lq
            zt_lq2 = -zt_lq*zt_lq
            ezt_lq2 = -np.exp(zt_lq2)
            ezgamt = np.exp(z_lq2 + gamt)
            ezgamct = np.exp(z_lq2 + gamct)

            # Upsilon calculations
            fullind = np.ix_(ind, index2)
            upsi1 = - np.exp(z_lq2 + gamct + np.log(wofz(1j*(z_lq + nuc[fullind])).real))
            tz = t1-z
            z1 = zt_lq + nuc[fullind]
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            if indv1[0].shape > 0:
                upsi1[indv1] += np.exp(zt_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                nuac2 = nuc[ind[indv2[0]], index2[indv2[1]]]**2
                upsi1[indv2] += np.exp(nuac2 - gamc[ind[indv2[0]], 0]*tz[indv2] + np.log(2.))\
                               - np.exp(zt_lq2[indv2] + np.log(wofz(-1j*z1[indv2]).real))
            upsi1[t1[:, 0] == 0., :] = 0.

            upsi2 = - np.exp(z_lq2 + gamt + np.log(wofz(1j*(z_lq + nu[fullind])).real))
            z1 = zt_lq + nu[fullind]
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            if indv1[0].shape > 0:
                upsi2[indv1] += np.exp(zt_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                nua2 = nu[ind[indv2[0]], index2[indv2[1]]]**2
                upsi2[indv2] += np.exp(nua2 - gam[ind[indv2[0]], 0]*tz[indv2] + np.log(2.))\
                               - np.exp(zt_lq2[indv2] + np.log(wofz(-1j*z1[indv2]).real))
            upsi2[t1[:, 0] == 0., :] = 0.

            #Gradient wrt lq
            la1 = S_wpi*nu2
            la1c = S_wpi*nuc2
            la2 = S_w*(.5*lq)
            uplq = ezt_lq2*(gamc_2[ind]) + ezgamct*(-z_lq/lq[0, index2] + gamc_2[ind])\
                - ezt_lq2*(gam_2[ind]) - ezgamt*(-z_lq/lq[0, index2] + gam_2[ind])

            glq[ind2t] = la1c[np.ix_(ind, index2)]*upsi1 - la1[np.ix_(ind, index2)]*upsi2\
                + la2[np.ix_(ind, index2)]*uplq


            #Gradient wrt S
            Sa1 = (lq*(-.25*np.sqrt(np.pi)))/w

            gSdq[ind2t] = Sa1[np.ix_(ind, index2)]*(upsi1 - upsi2)

            #Gradient wrt B
            #Dx1 terms
            dgam_dB = .5/w
            dgamc_dB = -dgam_dB
            #DxQ terms
            Ba1 = .5*(c0/w2)
            Ba2 = c0*dgam_dB
            Ba3 = lq2*gam_2
            Ba4 = (dgam_dB*S_w)*(.25*lq2)

            Ba2c = c0*dgamc_dB
            Ba3c = lq2*gamc_2
            Ba4c = (dgamc_dB*S_w)*(.25*lq2)

            gB[ind2t] = (Ba1[np.ix_(ind, index2)] + Ba2c[np.ix_(ind, index2)]*(Ba3c[np.ix_(ind, index2)] - (t1-z)))*upsi1\
                + Ba4c[np.ix_(ind, index2)]*(ezt_lq2 + ezgamct)\
                - (Ba1[np.ix_(ind, index2)] + Ba2[np.ix_(ind, index2)]*(Ba3[np.ix_(ind, index2)] - (t1-z)))*upsi2\
                - Ba4[np.ix_(ind, index2)]*(ezt_lq2 + ezgamt)

            #Gradient wrt C
            #Dx1 terms
            dgam_dC = 0.5 - .5*(alphad/w)
            dgamc_dC = 0.5 + .5*(alphad/w)
            #DxQ terms
            Ca1 = -c0*(.5*alphad/w2)
            Ca2 = c0*dgam_dC
            Ca4 = (dgam_dC*S_w)*(.25*lq2)

            Ca2c = c0*dgamc_dC
            Ca4c = (dgamc_dC*S_w)*(.25*lq2)

            gC[ind2t] = (Ca1[np.ix_(ind, index2)] + Ca2c[np.ix_(ind, index2)]*(Ba3c[np.ix_(ind, index2)] - (t1-z)))*upsi1\
                + Ca4c[np.ix_(ind, index2)]*(ezt_lq2 + ezgamct)\
                - (Ca1[np.ix_(ind, index2)] + Ca2[np.ix_(ind, index2)]*(Ba3[np.ix_(ind, index2)] - (t1-z)))*upsi2\
                - Ca4[np.ix_(ind, index2)]*(ezt_lq2 + ezgamt)

        return glq, gSdq, gB, gC

    #TODO: reduce memory usage
    def _gkfu_z(self, X, index, Z, index2): #Kfu(t,z)
        index = index.reshape(index.size,)
        #terms that move along t
        d = np.unique(index)
        B = self.B[d].values
        C = self.C[d].values
        S = self.W[d, :].values
        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]
        #Check where wd becomes complex
        wbool = C*C >= 4.*B
        wbool2 = wbool[index]
        ind2t = np.where(wbool2)
        ind3t = np.where(np.logical_not(wbool2))
        #t column
        t = X[:, 0].reshape(X.shape[0], 1)
        C = C.reshape(C.size, 1)
        B = B.reshape(B.size, 1)
        C2 = C*C
        alpha = .5*C
        #z row
        z = Z[:, 0].reshape(1, Z.shape[0])
        index2 = index2.reshape(index2.size,)
        lq = self.lengthscale.values.reshape((1, self.rank))

        #kfu = np.empty((t.size, z.size))
        gz = np.empty((t.size, z.size))
        indD = np.arange(B.size)
        #(1) when wd is real
        if np.any(np.logical_not(wbool)):
            #Indexes of index and t related to (2)
            t1 = t[ind3t]
            ind = index[ind3t]
            #TODO: Find a better way of doing this
            #Index transformation
            d = np.asarray(np.where(np.logical_not(wbool))[0])
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            w = .5*np.sqrt(4.*B[d] - C2[d])
            alphad = alpha[d]
            gam = alphad - 1j*w
            S_w = S[d]/w
            S_wpi =S_w*(.5*np.sqrt(np.pi))
            #DxQ terms
            c0 = S_wpi*lq #lq*Sdq*sqrt(pi)/(2w)
            nu = (.5*gam)*lq

            #1xM terms
            z_lq = z/lq[0, index2]
            z_lq2 = -z_lq*z_lq
            #NxQ terms
            t_lq = t1/lq
            #DxM terms
            gamt = -gam[ind]*t1
            #NxM terms
            zt_lq = z_lq - t_lq[:, index2]
            zt_lq2 = -zt_lq*zt_lq
            #ezt_lq2 = -np.exp(zt_lq2)
            ezgamt = np.exp(z_lq2 + gamt)

            # Upsilon calculations
            fullind = np.ix_(ind, index2)
            upsi = - np.exp(z_lq2 + gamt + np.log(wofz(1j*(z_lq + nu[fullind]))))
            tz = t1-z
            z1 = zt_lq + nu[fullind]
            indv1 = np.where(z1.real >= 0.)
            indv2 = np.where(z1.real < 0.)
            if indv1[0].shape > 0:
                upsi[indv1] += np.exp(zt_lq2[indv1] + np.log(wofz(1j*z1[indv1])))
            if indv2[0].shape > 0:
                nua2 = nu[ind[indv2[0]], index2[indv2[1]]]**2
                upsi[indv2] += np.exp(nua2 - gam[ind[indv2[0]], 0]*tz[indv2] + np.log(2.))\
                               - np.exp(zt_lq2[indv2] + np.log(wofz(-1j*z1[indv2])))
            upsi[t1[:, 0] == 0., :] = 0.

            #Gradient wrt z
            za1 = c0*gam
            #za2 = S_w
            gz[ind3t] = (za1[np.ix_(ind, index2)]*upsi).imag + S_w[np.ix_(ind, index2)]*ezgamt.imag

        #(2) when wd is complex
        if np.any(wbool):
            #Indexes of index and t related to (2)
            t1 = t[ind2t]
            ind = index[ind2t]
            #Index transformation
            d = np.asarray(np.where(wbool)[0])
            indd = indD.copy()
            indd[d] = np.arange(d.size)
            ind = indd[ind]
            #Dx1 terms
            w = .5*np.sqrt(C2[d] - 4.*B[d])
            alphad = alpha[d]
            gam = alphad - w
            gamc = alphad + w
            #DxQ terms
            S_w = -S[d]/w #minus is given by j*j
            S_wpi = S_w*(.25*np.sqrt(np.pi))
            c0 = S_wpi*lq
            nu = .5*gam*lq
            nuc = .5*gamc*lq

            #1xM terms
            z_lq = z/lq[0, index2]
            z_lq2 = -z_lq*z_lq
            #Nx1
            gamt = -gam[ind]*t1
            gamct = -gamc[ind]*t1
            #NxQ terms
            t_lq = t1/lq
            #NxM terms
            zt_lq = z_lq - t_lq[:, index2]
            ezgamt = np.exp(z_lq2 + gamt)
            ezgamct = np.exp(z_lq2 + gamct)

            # Upsilon calculations
            zt_lq2 = -zt_lq*zt_lq
            fullind = np.ix_(ind, index2)
            upsi1 = - np.exp(z_lq2 + gamct + np.log(wofz(1j*(z_lq + nuc[fullind])).real))
            tz = t1-z
            z1 = zt_lq + nuc[fullind]
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            if indv1[0].shape > 0:
                upsi1[indv1] += np.exp(zt_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                nuac2 = nuc[ind[indv2[0]], index2[indv2[1]]]**2
                upsi1[indv2] += np.exp(nuac2 - gamc[ind[indv2[0]], 0]*tz[indv2] + np.log(2.))\
                               - np.exp(zt_lq2[indv2] + np.log(wofz(-1j*z1[indv2]).real))
            upsi1[t1[:, 0] == 0., :] = 0.

            upsi2 = - np.exp(z_lq2 + gamt + np.log(wofz(1j*(z_lq + nu[fullind])).real))
            z1 = zt_lq + nu[fullind]
            indv1 = np.where(z1 >= 0.)
            indv2 = np.where(z1 < 0.)
            if indv1[0].shape > 0:
                upsi2[indv1] += np.exp(zt_lq2[indv1] + np.log(wofz(1j*z1[indv1]).real))
            if indv2[0].shape > 0:
                nua2 = nu[ind[indv2[0]], index2[indv2[1]]]**2
                upsi2[indv2] += np.exp(nua2 - gam[ind[indv2[0]], 0]*tz[indv2] + np.log(2.))\
                               - np.exp(zt_lq2[indv2] + np.log(wofz(-1j*z1[indv2]).real))
            upsi2[t1[:, 0] == 0., :] = 0.

            #Gradient wrt z
            za1 = c0*gam
            za1c = c0*gamc
            za2 = .5*S_w
            gz[ind2t] = za1c[np.ix_(ind, index2)]*upsi1 - za1[np.ix_(ind, index2)]*upsi2\
                + za2[np.ix_(ind, index2)]*(ezgamct - ezgamt)
        return gz



# Copyright (c) 2014, Cristian Guarnizo.
# Licensed under the BSD 3-clause license (see LICENSE.txt)

import numpy as np
from scipy.special import erf, erfcx
from .kern import Kern
from ...core.parameterization import Param
from paramz.transformations import Logexp
from paramz.caching import Cache_this

class EQ_ODE1(Kern):
    """
    Covariance function for first order differential equation driven by an exponentiated quadratic covariance.

    This outputs of this kernel have the form
    .. math::
       \frac{\text{d}y_j}{\text{d}t} = \sum_{i=1}^R w_{j,i} u_i(t-\delta_j) - d_jy_j(t)

    where :math:`R` is the rank of the system, :math:`w_{j,i}` is the sensitivity of the :math:`j`th output to the :math:`i`th latent function, :math:`d_j` is the decay rate of the :math:`j`th output and :math:`u_i(t)` are independent latent Gaussian processes goverened by an exponentiated quadratic covariance.
    
    :param output_dim: number of outputs driven by latent function.
    :type output_dim: int
    :param W: sensitivities of each output to the latent driving function. 
    :type W: ndarray (output_dim x rank).
    :param rank: If rank is greater than 1 then there are assumed to be a total of rank latent forces independently driving the system, each with identical covariance.
    :type rank: int
    :param decay: decay rates for the first order system. 
    :type decay: array of length output_dim.
    :param delay: delay between latent force and output response.
    :type delay: array of length output_dim.
    :param kappa: diagonal term that allows each latent output to have an independent component to the response.
    :type kappa: array of length output_dim.
    
    .. Note: see first order differential equation examples in GPy.examples.regression for some usage.
    """
    def __init__(self, input_dim=2, output_dim=1, rank=1, W = None, lengthscale=None,  decay=None, active_dims=None, name='eq_ode1'):
        assert input_dim == 2, "only defined for 1 input dims"
        super(EQ_ODE1, self).__init__(input_dim=input_dim, active_dims=active_dims, name=name)

        self.rank = rank
        self.output_dim = output_dim

        if lengthscale is None:
            lengthscale = .5 + np.random.rand(self.rank)
        else:
            lengthscale = np.asarray(lengthscale)
            assert lengthscale.size in [1, self.rank], "Bad number of lengthscales"
            if lengthscale.size != self.rank:
                lengthscale = np.ones(self.rank)*lengthscale
            
        if W is None:
            W = .5*np.random.randn(self.output_dim, self.rank)/np.sqrt(self.rank)
        else:
            assert W.shape == (self.output_dim, self.rank)
        
        if decay is None:
            decay = np.ones(self.output_dim)
        else:
            decay = np.asarray(decay)
            assert decay.size in [1, self.output_dim], "Bad number of decay"
            if decay.size != self.output_dim:
                decay = np.ones(self.output_dim)*decay

#        if kappa is None:
#            self.kappa = np.ones(self.output_dim)
#        else:
#            kappa = np.asarray(kappa)
#            assert kappa.size in [1, self.output_dim], "Bad number of kappa"
#            if decay.size != self.output_dim:
#                decay = np.ones(self.output_dim)*kappa

        #self.kappa = Param('kappa', kappa, Logexp())
        #self.delay = Param('delay', delay, Logexp())
        #self.is_normalized = True
        #self.is_stationary = False
        #self.gaussian_initial = False

        self.lengthscale = Param('lengthscale', lengthscale, Logexp())
        self.decay = Param('decay', decay, Logexp())
        self.W = Param('W', W)
        self.link_parameters(self.lengthscale, self.decay, self.W)

    @Cache_this(limit=3)
    def K(self, X, X2=None):
        #This way is not working, indexes are lost after using k._slice_X
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        index = np.int_(np.round(X[:, 1]))
        index = index.reshape(index.size,)
        X_flag = index[0] >= self.output_dim
        if X2 is None:
            if X_flag:
                #Calculate covariance function for the latent functions
                index -= self.output_dim
                return self._Kuu(X, index)
            else:
                raise NotImplementedError
        else:
            #This way is not working, indexes are lost after using k._slice_X
            #index2 = np.asarray(X2, dtype=np.int)
            #index2 = index2.reshape(index2.size,)
            if hasattr(X2, 'values'):
                X2 = X2.values
            index2 = np.int_(np.round(X2[:, 1]))
            index2 = index2.reshape(index2.size,)
            X2_flag = index2[0] >= self.output_dim
            #Calculate cross-covariance function
            if not X_flag and X2_flag:
                index2 -= self.output_dim
                return self._Kfu(X, index, X2, index2) #Kfu
            elif X_flag and not X2_flag:
                index -= self.output_dim
                return self._Kfu(X2, index2, X, index).T #Kuf
            elif X_flag and X2_flag:
                index -= self.output_dim
                index2 -= self.output_dim
                return self._Kusu(X, index, X2, index2) #Ku_s u
            else:
                raise NotImplementedError #Kf_s f

    #Calculate the covariance function for diag(Kff(X,X))
    def Kdiag(self, X):
        if hasattr(X, 'values'):
            index = np.int_(np.round(X[:, 1].values))
        else:
            index = np.int_(np.round(X[:, 1]))
        index = index.reshape(index.size,)
        X_flag = index[0] >= self.output_dim
        
        if X_flag: #Kuudiag        
            return np.ones(X[:,0].shape)
        else: #Kffdiag
            kdiag = self._Kdiag(X)
            return np.sum(kdiag, axis=1)
        
    def _Kdiag(self, X):
        #This way is not working, indexes are lost after using k._slice_X
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        index = np.int_(X[:, 1])
        index = index.reshape(index.size,)
        
        #terms that move along t
        t = X[:, 0].reshape(X.shape[0], 1)
        d = np.unique(index) #Output Indexes
        B = self.decay.values[d]
        S = self.W.values[d, :]
        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]
        
        B = B.reshape(B.size, 1)
        #Terms that move along q
        lq = self.lengthscale.values.reshape(1, self.rank)
        S2 = S*S
        kdiag = np.empty((t.size, ))

        #Dx1 terms
        c0 = (S2/B)*((.5*np.sqrt(np.pi))*lq)

        #DxQ terms
        nu = lq*(B*.5)
        nu2 = nu*nu
        #Nx1 terms
        gamt = -2.*B
        gamt = gamt[index]*t

        #NxQ terms
        t_lq = t/lq

        # Upsilon Calculations
        # Using wofz
        #erfnu = erf(nu)
        
        upm = np.exp(nu2[index, :] + lnDifErf( nu[index, :] ,t_lq+nu[index,:] ))
        upm[t[:, 0] == 0, :] = 0.

        
        upv = np.exp(nu2[index, :] + gamt + lnDifErf( -t_lq+nu[index,:], nu[index, :] ) )
        upv[t[:, 0] == 0, :] = 0.

        #Covariance calculation
        #kdiag = np.sum(c0[index, :]*(upm-upv), axis=1)
        kdiag = c0[index, :]*(upm-upv)
        return kdiag

    def update_gradients_full(self, dL_dK, X, X2 = None):
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        self.decay.gradient = np.zeros(self.decay.shape)
        self.W.gradient = np.zeros(self.W.shape)
        self.lengthscale.gradient = np.zeros(self.lengthscale.shape)
        index = np.int_(np.round(X[:, 1]))
        index = index.reshape(index.size,)
        X_flag = index[0] >= self.output_dim
        if X2 is None:
            if X_flag: #Kuu or Kmm
                index -= self.output_dim
                tmp = dL_dK*self._gkuu_lq(X, index)
                for q in np.unique(index):
                    ind = np.where(index == q)
                    self.lengthscale.gradient[q] = tmp[np.ix_(ind[0], ind[0])].sum()
            else:
                raise NotImplementedError
        else: #Kfu or Knm
            #index2 = np.asarray(X2, dtype=np.int)
            #index2 = index2.reshape(index2.size,)
            if hasattr(X2, 'values'):
                X2 = X2.values
            index2 = np.int_(np.round(X2[:, 1]))
            index2 = index2.reshape(index2.size,)
            X2_flag = index2[0] >= self.output_dim
            if not X_flag and X2_flag: #Kfu
                index2 -= self.output_dim
            else: #Kuf
                dL_dK = dL_dK.T #so we obtaing dL_Kfu
                indtemp = index - self.output_dim
                Xtemp = X
                X = X2
                X2 = Xtemp
                index = index2
                index2 = indtemp
            glq, gSdq, gB = self._gkfu(X, index, X2, index2)
            tmp = dL_dK*glq
            for q in np.unique(index2):
                ind = np.where(index2 == q)
                self.lengthscale.gradient[q] = tmp[:, ind].sum()
            tmpB = dL_dK*gB
            tmp = dL_dK*gSdq
            for d in np.unique(index):
                ind = np.where(index == d)
                self.decay.gradient[d] = tmpB[ind, :].sum()
                for q in np.unique(index2):
                    ind2 = np.where(index2 == q)
                    self.W.gradient[d, q] = tmp[np.ix_(ind[0], ind2[0])].sum()

    def update_gradients_diag(self, dL_dKdiag, X):
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        self.decay.gradient = np.zeros(self.decay.shape)
        self.W.gradient = np.zeros(self.W.shape)
        self.lengthscale.gradient = np.zeros(self.lengthscale.shape)
        index = np.int_(X[:, 1])
        index = index.reshape(index.size,)
        
        glq, gS, gB = self._gkdiag(X, index)
        if dL_dKdiag.size == X.shape[0]:
            dL_dKdiag = np.reshape(dL_dKdiag, (index.size, 1))
        tmp = dL_dKdiag*glq
        self.lengthscale.gradient = tmp.sum(0)
        tmpB = dL_dKdiag*gB
        tmp = dL_dKdiag*gS
        for d in np.unique(index):
            ind = np.where(index == d)
            self.decay.gradient[d] = tmpB[ind, :].sum()
            self.W.gradient[d, :] = tmp[ind].sum(0)

    def gradients_X(self, dL_dK, X, X2=None):
        #index = np.asarray(X, dtype=np.int)
        #index = index.reshape(index.size,)
        if hasattr(X, 'values'):
            X = X.values
        index = np.int_(np.round(X[:, 1]))
        index = index.reshape(index.size,)
        X_flag = index[0] >= self.output_dim
        #If input_dim == 1, use this
        #gX = np.zeros((X.shape[0], 1))
        #Cheat to allow gradient for input_dim==2
        gX = np.zeros(X.shape)
        if X2 is None: #Kuu or Kmm
            if X_flag:
                index -= self.output_dim
                gX[:, 0] = 2.*(dL_dK*self._gkuu_X(X, index)).sum(0)
                return gX
            else:
                raise NotImplementedError
        else: #Kuf or Kmn
            #index2 = np.asarray(X2, dtype=np.int)
            #index2 = index2.reshape(index2.size,)
            if hasattr(X2, 'values'):
                X2 = X2.values
            index2 = np.int_(np.round(X2[:, 1]))
            index2 = index2.reshape(index2.size,)
            X2_flag = index2[0] >= self.output_dim
            if X_flag and not X2_flag: #gradient of Kuf(Z, X) wrt Z
                index -= self.output_dim
                gX[:, 0] = (dL_dK*self._gkfu_z(X2, index2, X, index).T).sum(1)
                return gX
            else:
                raise NotImplementedError

    #---------------------------------------#
    #             Helper functions          #
    #---------------------------------------#

    #Evaluation of squared exponential for LFM
    def _Kuu(self, X, index):
        index = index.reshape(index.size,)
        t = X[:, 0].reshape(X.shape[0],)
        lq = self.lengthscale.values.reshape(self.rank,)
        lq2 = lq*lq
        #Covariance matrix initialization
        kuu = np.zeros((t.size, t.size))
        #Assign 1. to diagonal terms
        kuu[np.diag_indices(t.size)] = 1.
        #Upper triangular indices
        indtri1, indtri2 = np.triu_indices(t.size, 1)
        #Block Diagonal indices among Upper Triangular indices
        ind = np.where(index[indtri1] == index[indtri2])
        indr = indtri1[ind]
        indc = indtri2[ind]
        r = t[indr] - t[indc]
        r2 = r*r
        #Calculation of  covariance function
        kuu[indr, indc] = np.exp(-r2/lq2[index[indr]])
        #Completion of lower triangular part
        kuu[indc, indr] = kuu[indr, indc]
        return kuu

    def _Kusu(self, X, index, X2, index2):
        index = index.reshape(index.size,)
        index2 = index2.reshape(index2.size,)
        t = X[:, 0].reshape(X.shape[0],1)
        t2 = X2[:, 0].reshape(1,X2.shape[0])
        lq = self.lengthscale.values.reshape(self.rank,)
        #Covariance matrix initialization
        kuu = np.zeros((t.size, t2.size))
        for q in range(self.rank):
            ind1 = index == q
            ind2 = index2 == q
            r = t[ind1]/lq[q] - t2[0,ind2]/lq[q]
            r2 = r*r
            #Calculation of  covariance function
            kuu[np.ix_(ind1, ind2)] = np.exp(-r2)
        return kuu

    #Evaluation of cross-covariance function
    def _Kfu(self, X, index, X2, index2):
        #terms that move along t
        t = X[:, 0].reshape(X.shape[0], 1)
        d = np.unique(index) #Output Indexes
        B = self.decay.values[d]
        S = self.W.values[d, :]
        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]
        #Output related variables must be column-wise
        B = B.reshape(B.size, 1)
        #Input related variables must be row-wise
        z = X2[:, 0].reshape(1, X2.shape[0])
        lq = self.lengthscale.values.reshape((1, self.rank))

        kfu = np.empty((t.size, z.size))

        #DxQ terms
        c0 = S*((.5*np.sqrt(np.pi))*lq)
        nu = B*(.5*lq)
        nu2 = nu**2
        #1xM terms
        z_lq = z/lq[0, index2]
        #NxM terms
        tz = t-z
        tz_lq = tz/lq[0, index2]

        # Upsilon Calculations
        fullind = np.ix_(index, index2)

        upsi = np.exp(nu2[fullind] - B[index]*tz + lnDifErf( -tz_lq + nu[fullind], z_lq+nu[fullind]))
        upsi[t[:, 0] == 0, :] = 0.
        #Covariance calculation
        kfu = c0[fullind]*upsi

        return kfu

    #Gradient of Kuu wrt lengthscale
    def _gkuu_lq(self, X, index):
        t = X[:, 0].reshape(X.shape[0],)
        index = index.reshape(X.shape[0],)
        lq = self.lengthscale.values.reshape(self.rank,)
        lq2 = lq*lq
        #Covariance matrix initialization
        glq = np.zeros((t.size, t.size))
        #Upper triangular indices
        indtri1, indtri2 = np.triu_indices(t.size, 1)
        #Block Diagonal indices among Upper Triangular indices
        ind = np.where(index[indtri1] == index[indtri2])
        indr = indtri1[ind]
        indc = indtri2[ind]
        r = t[indr] - t[indc]
        r2 = r*r
        r2_lq2 = r2/lq2[index[indr]]
        #Calculation of  covariance function
        er2_lq2 = np.exp(-r2_lq2)
        #Gradient wrt lq
        c = 2.*r2_lq2/lq[index[indr]]
        glq[indr, indc] = er2_lq2*c
        #Complete the lower triangular
        glq[indc, indr] = glq[indr, indc]
        return glq

    #Be careful this derivative should be transpose it
    def _gkuu_X(self, X, index): #Diagonal terms are always zero
        t = X[:, 0].reshape(X.shape[0],)
        index = index.reshape(index.size,)
        lq = self.lengthscale.values.reshape(self.rank,)
        lq2 = lq*lq
        #Covariance matrix initialization
        gt = np.zeros((t.size, t.size))
        #Upper triangular indices
        indtri1, indtri2 = np.triu_indices(t.size, 1) #Offset of 1 from the diagonal
        #Block Diagonal indices among Upper Triangular indices
        ind = np.where(index[indtri1] == index[indtri2])
        indr = indtri1[ind]
        indc = indtri2[ind]
        r = t[indr] - t[indc]
        r2 = r*r
        r2_lq2 = r2/(-lq2[index[indr]])
        #Calculation of  covariance function
        er2_lq2 = np.exp(r2_lq2)
        #Gradient wrt t
        c = 2.*r/lq2[index[indr]]
        gt[indr, indc] = er2_lq2*c
        #Complete the lower triangular
        gt[indc, indr] = -gt[indr, indc]
        return gt

    #Gradients for Diagonal Kff
    def _gkdiag(self, X, index):
        index = index.reshape(index.size,)
        #terms that move along t
        d = np.unique(index)
        B = self.decay[d].values
        S = self.W[d, :].values
        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]
        #Output related variables must be column-wise
        t = X[:, 0].reshape(X.shape[0], 1)
        B = B.reshape(B.size, 1)
        S2 = S*S

        #Input related variables must be row-wise
        lq = self.lengthscale.values.reshape(1, self.rank)

        gB = np.empty((t.size,))
        glq = np.empty((t.size, lq.size))
        gS = np.empty((t.size, lq.size))

        #Dx1 terms
        c0 = S2*lq*np.sqrt(np.pi)

        #DxQ terms
        nu = (.5*lq)*B
        nu2 = nu*nu
        
        #Nx1 terms
        gamt = -B[index]*t
        egamt = np.exp(gamt)
        e2gamt = egamt*egamt

        #NxQ terms
        t_lq = t/lq
        t2_lq2 = -t_lq*t_lq

        etlq2gamt = np.exp(t2_lq2 + gamt) #NXQ

        ##Upsilon calculations
        #erfnu = erf(nu) #TODO: This can be improved

        upm = np.exp(nu2[index, :] + lnDifErf( nu[index, :], t_lq + nu[index, :]) )
        upm[t[:, 0] == 0, :] = 0.

        upv = np.exp(nu2[index, :] + 2.*gamt + lnDifErf(-t_lq + nu[index, :], nu[index, :]) ) #egamt*upv
        upv[t[:, 0] == 0, :] = 0.

        #Gradient wrt S
        c0_S = (S/B)*(lq*np.sqrt(np.pi))

        gS = c0_S[index]*(upm - upv)

        #For B
        CB1 = (.5*lq)**2 - .5/B**2 #DXQ
        lq2_2B = (.5*lq**2)*(S2/B) #DXQ
        CB2 = 2.*etlq2gamt - e2gamt - 1. #NxQ
        
        # gradient wrt B NxZ
        gB = c0[index, :]*(CB1[index, :]*upm - (CB1[index, :] - t/B[index])*upv) + \
        lq2_2B[index, :]*CB2

        #Gradient wrt lengthscale
        #DxQ terms
        c0 = (.5*np.sqrt(np.pi))*(S2/B)*(1.+.5*(lq*B)**2)
        Clq1 = S2*(lq*.5)
        glq = c0[index]*(upm - upv) + Clq1[index]*CB2

        return glq, gS, gB

    def _gkfu(self, X, index, Z, index2):
        index = index.reshape(index.size,)
        #TODO: reduce memory usage
        #terms that move along t
        d = np.unique(index)
        B = self.decay[d].values
        S = self.W[d, :].values

        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]
        #t column
        t = X[:, 0].reshape(X.shape[0], 1)
        B = B.reshape(B.size, 1)
        #z row
        z = Z[:, 0].reshape(1, Z.shape[0])
        index2 = index2.reshape(index2.size,)
        lq = self.lengthscale.values.reshape((1, self.rank))

        #kfu = np.empty((t.size, z.size))
        glq = np.empty((t.size, z.size))
        gSdq = np.empty((t.size, z.size))
        gB = np.empty((t.size, z.size))

        #Dx1 terms
        B_2 = B*.5
        S_pi = S*(.5*np.sqrt(np.pi))
        #DxQ terms
        c0 = S_pi*lq #lq*Sdq*sqrt(pi)
        nu = B*lq*.5
        nu2 = nu*nu

        #1xM terms
        z_lq = z/lq[0, index2]
        
        #NxM terms
        tz = t-z
        tz_lq = tz/lq[0, index2]
        etz_lq2 = -np.exp(-tz_lq*tz_lq)
        ez_lq_Bt = np.exp(-z_lq*z_lq -B[index]*t)
        
        # Upsilon calculations
        fullind = np.ix_(index, index2)
        upsi = np.exp(nu2[fullind] - B[index]*tz + lnDifErf( -tz_lq + nu[fullind], z_lq+nu[fullind] ) )
        upsi[t[:, 0] == 0., :] = 0.

        #Gradient wrt S
        #DxQ term
        Sa1 = lq*(.5*np.sqrt(np.pi))

        gSdq = Sa1[0,index2]*upsi

        #Gradient wrt lq
        la1 = S_pi*(1. + 2.*nu2)
        Slq = S*lq
        uplq = etz_lq2*(tz_lq/lq[0, index2] + B_2[index])
        uplq += ez_lq_Bt*(-z_lq/lq[0, index2] + B_2[index])

        glq = la1[fullind]*upsi
        glq += Slq[fullind]*uplq

        #Gradient wrt B
        Slq = Slq*lq
        nulq = nu*lq
        upBd = etz_lq2 + ez_lq_Bt
        gB = c0[fullind]*(nulq[fullind] - tz)*upsi + .5*Slq[fullind]*upBd

        return glq, gSdq, gB

    #TODO: reduce memory usage
    def _gkfu_z(self, X, index, Z, index2): #Kfu(t,z)
        index = index.reshape(index.size,)
        #terms that move along t
        d = np.unique(index)
        B = self.decay[d].values
        S = self.W[d, :].values
        #Index transformation
        indd = np.arange(self.output_dim)
        indd[d] = np.arange(d.size)
        index = indd[index]

        #t column
        t = X[:, 0].reshape(X.shape[0], 1)
        B = B.reshape(B.size, 1)
        #z row
        z = Z[:, 0].reshape(1, Z.shape[0])
        index2 = index2.reshape(index2.size,)
        lq = self.lengthscale.values.reshape((1, self.rank))

        #kfu = np.empty((t.size, z.size))
        gz = np.empty((t.size, z.size))

        #Dx1 terms
        S_pi =S*(.5*np.sqrt(np.pi))
        #DxQ terms
        #Slq = S*lq
        c0 = S_pi*lq #lq*Sdq*sqrt(pi)
        nu = (.5*lq)*B
        nu2 = nu*nu

        #1xM terms
        z_lq = z/lq[0, index2]
        z_lq2 = -z_lq*z_lq
        #NxQ terms
        t_lq = t/lq
        #NxM terms
        zt_lq = z_lq - t_lq[:, index2]
        zt_lq2 = -zt_lq*zt_lq

        # Upsilon calculations
        fullind = np.ix_(index, index2)
        z2 = z_lq + nu[fullind]
        z1 = z2 - t_lq[:, index2]
        upsi = np.exp(nu2[fullind] - B[index]*(t-z) + lnDifErf(z1,z2) )
        upsi[t[:, 0] == 0., :] = 0.

        #Gradient wrt z
        za1 = c0*B
        #za2 = S_w
        gz = za1[fullind]*upsi + S[fullind]*( np.exp(z_lq2 - B[index]*t) -np.exp(zt_lq2) )

        return gz
        
def lnDifErf(z1,z2):
    #Z2 is always positive
    logdiferf = np.zeros(z1.shape)        
    ind = np.where(z1>0.)
    ind2 = np.where(z1<=0.)
    if ind[0].shape > 0:
        z1i = z1[ind]
        z12 = z1i*z1i
        z2i = z2[ind]
        logdiferf[ind] = -z12 + np.log(erfcx(z1i) - erfcx(z2i)*np.exp(z12-z2i**2))
    
    if ind2[0].shape > 0:
        z1i = z1[ind2]
        z2i = z2[ind2]
        logdiferf[ind2] = np.log(erf(z2i) - erf(z1i))
        
    return logdiferf


# Copyright (c) 2012, GPy authors (see AUTHORS.txt).
# Licensed under the BSD 3-clause license (see LICENSE.txt)

from .posterior import Posterior
from ...util.linalg import mdot, jitchol, backsub_both_sides, tdot, dtrtrs, dtrtri, dpotri, dpotrs, symmetrify
from ...util import diag
from GPy.core.parameterization.variational import VariationalPosterior
import numpy as np
from . import LatentFunctionInference
log_2_pi = np.log(2*np.pi)
import logging, itertools
logger = logging.getLogger('vardtc')

class VarDTC(LatentFunctionInference):
    """
    An object for inference when the likelihood is Gaussian, but we want to do sparse inference.

    The function self.inference returns a Posterior object, which summarizes
    the posterior.

    For efficiency, we sometimes work with the cholesky of Y*Y.T. To save repeatedly recomputing this, we cache it.

    """
    const_jitter = 1e-8
    def __init__(self, limit=1):
        from paramz.caching import Cacher
        self.limit = limit
        self.get_trYYT = Cacher(self._get_trYYT, limit)
        self.get_YYTfactor = Cacher(self._get_YYTfactor, limit)

    def set_limit(self, limit):
        self.get_trYYT.limit = limit
        self.get_YYTfactor.limit = limit

    def _get_trYYT(self, Y):
        return np.einsum("ij,ij->", Y, Y)
        # faster than, but same as:
        # return np.sum(np.square(Y))

    def __getstate__(self):
        # has to be overridden, as Cacher objects cannot be pickled.
        return self.limit

    def __setstate__(self, state):
        # has to be overridden, as Cacher objects cannot be pickled.
        self.limit = state
        from paramz.caching import Cacher
        self.get_trYYT = Cacher(self._get_trYYT, self.limit)
        self.get_YYTfactor = Cacher(self._get_YYTfactor, self.limit)

    def _get_YYTfactor(self, Y):
        """
        find a matrix L which satisfies LLT = YYT.

        Note that L may have fewer columns than Y.
        """
        N, D = Y.shape
        if (N>=D):
            return Y.view(np.ndarray)
        else:
            return jitchol(tdot(Y))

    def get_VVTfactor(self, Y, prec):
        return Y * prec # TODO cache this, and make it effective

    def inference(self, kern, X, Z, likelihood, Y, Y_metadata=None, mean_function=None, precision=None, Lm=None, dL_dKmm=None, psi0=None, psi1=None, psi2=None, Z_tilde=None):

        num_data, output_dim = Y.shape
        num_inducing = Z.shape[0]

        uncertain_inputs = isinstance(X, VariationalPosterior)

        if mean_function is not None:
            mean = mean_function.f(X)
        else:
            mean = 0

        if precision is None:
            #assume Gaussian likelihood
            precision = 1./np.fmax(likelihood.gaussian_variance(Y_metadata), self.const_jitter)

        if precision.ndim == 1:
            precision = precision[:, None]
        het_noise = precision.size > 1
        if (het_noise or uncertain_inputs) and mean_function is not None:
            raise ValueError('Mean function not implemented with uncertain inputs or heteroscedasticity')

        VVT_factor = precision*(Y-mean)
        trYYT = self.get_trYYT(Y-mean)

        # kernel computations, using BGPLVM notation
        if Lm is None:
            Kmm = kern.K(Z).copy()
            diag.add(Kmm, self.const_jitter)
            Lm = jitchol(Kmm)
        else:
            Kmm = tdot(Lm)
            symmetrify(Kmm)


        # The rather complex computations of A, and the psi stats
        if uncertain_inputs:
            if psi0 is None:
                psi0 = kern.psi0(Z, X)
            if psi1 is None:
                psi1 = kern.psi1(Z, X)
            if het_noise:
                if psi2 is None:
                    psi2_beta = (kern.psi2n(Z, X) * precision[:, :, None]).sum(0)
                else:
                    psi2_beta = (psi2 * precision[:, :, None]).sum(0)
            else:
                if psi2 is None:
                    psi2_beta = kern.psi2(Z,X) * precision
                elif psi2.ndim == 3:
                    psi2_beta = psi2.sum(0) * precision
                else:
                    psi2_beta = psi2 * precision
            LmInv = dtrtri(Lm)
            A = LmInv.dot(psi2_beta.dot(LmInv.T))
        else:
            if psi0 is None:
                psi0 = kern.Kdiag(X)
            if psi1 is None:
                psi1 = kern.K(X, Z)
            if het_noise:
                tmp = psi1 * (np.sqrt(precision))
            else:
                tmp = psi1 * (np.sqrt(precision))
            tmp, _ = dtrtrs(Lm, tmp.T, lower=1)
            A = tdot(tmp) #print A.sum()

        # factor B
        B = np.eye(num_inducing) + A
        LB = jitchol(B)

        # back substutue C into psi1Vf
        tmp, _ = dtrtrs(Lm, psi1.T, lower=1, trans=0)
        _LBi_Lmi_psi1, _ = dtrtrs(LB, tmp, lower=1, trans=0)
        _LBi_Lmi_psi1Vf = np.dot(_LBi_Lmi_psi1, VVT_factor)
        tmp, _ = dtrtrs(LB, _LBi_Lmi_psi1Vf, lower=1, trans=1)
        Cpsi1Vf, _ = dtrtrs(Lm, tmp, lower=1, trans=1)

        # data fit and derivative of L w.r.t. Kmm
        dL_dm = -_LBi_Lmi_psi1.T.dot(_LBi_Lmi_psi1.dot(VVT_factor)) + VVT_factor

        delit = tdot(_LBi_Lmi_psi1Vf)
        data_fit = np.trace(delit)
        DBi_plus_BiPBi = backsub_both_sides(LB, output_dim * np.eye(num_inducing) + delit)
        if dL_dKmm is None:
            delit = -0.5 * DBi_plus_BiPBi
            delit += -0.5 * B * output_dim
            delit += output_dim * np.eye(num_inducing)
            # Compute dL_dKmm
            dL_dKmm = backsub_both_sides(Lm, delit)

        # derivatives of L w.r.t. psi
        dL_dpsi0, dL_dpsi1, dL_dpsi2 = _compute_dL_dpsi(num_inducing, num_data, output_dim, precision, Lm,
            VVT_factor, Cpsi1Vf, DBi_plus_BiPBi,
            psi1, het_noise, uncertain_inputs)

        # log marginal likelihood
        log_marginal = _compute_log_marginal_likelihood(likelihood, num_data, output_dim, precision, het_noise,
            psi0, A, LB, trYYT, data_fit, Y)

        if Z_tilde is not None:
            # This is a correction term for the log marginal likelihood
            # In EP this is log Z_tilde, which is the difference between the
            # Gaussian marginal and Z_EP
            log_marginal += Z_tilde

        #noise derivatives
        dL_dR = _compute_dL_dR(likelihood,
            het_noise, uncertain_inputs, LB,
            _LBi_Lmi_psi1Vf, DBi_plus_BiPBi, Lm, A,
            psi0, psi1, precision,
            data_fit, num_data, output_dim, trYYT, Y, VVT_factor)

        dL_dthetaL = likelihood.exact_inference_gradients(dL_dR,Y_metadata)

        #put the gradients in the right places
        if uncertain_inputs:
            grad_dict = {'dL_dKmm': dL_dKmm,
                         'dL_dpsi0':dL_dpsi0,
                         'dL_dpsi1':dL_dpsi1,
                         'dL_dpsi2':dL_dpsi2,
                         'dL_dthetaL':dL_dthetaL}
        else:
            grad_dict = {'dL_dKmm': dL_dKmm,
                         'dL_dKdiag':dL_dpsi0,
                         'dL_dKnm':dL_dpsi1,
                         'dL_dthetaL':dL_dthetaL,
                         'dL_dm':dL_dm}

        #get sufficient things for posterior prediction
        #TODO: do we really want to do this in  the loop?
        if VVT_factor.shape[1] == Y.shape[1]:
            woodbury_vector = Cpsi1Vf # == Cpsi1V
        else:
            print('foobar')
            import ipdb; ipdb.set_trace()
            psi1V = np.dot(Y.T*precision, psi1).T
            tmp, _ = dtrtrs(Lm, psi1V, lower=1, trans=0)
            tmp, _ = dpotrs(LB, tmp, lower=1)
            woodbury_vector, _ = dtrtrs(Lm, tmp, lower=1, trans=1)
        #Bi, _ = dpotri(LB, lower=1)
        #symmetrify(Bi)
        Bi = -dpotri(LB, lower=1)[0]
        diag.add(Bi, 1)

        woodbury_inv = backsub_both_sides(Lm, Bi)

        #construct a posterior object
        post = Posterior(woodbury_inv=woodbury_inv, woodbury_vector=woodbury_vector, K=Kmm, mean=None, cov=None, K_chol=Lm)
        return post, log_marginal, grad_dict

def _compute_dL_dpsi(num_inducing, num_data, output_dim, beta, Lm, VVT_factor, Cpsi1Vf, DBi_plus_BiPBi, psi1, het_noise, uncertain_inputs):
    dL_dpsi0 = -0.5 * output_dim * (beta* np.ones([num_data, 1])).flatten()
    dL_dpsi1 = np.dot(VVT_factor, Cpsi1Vf.T)
    dL_dpsi2_beta = 0.5 * backsub_both_sides(Lm, output_dim * np.eye(num_inducing) - DBi_plus_BiPBi)
    if het_noise:
        if uncertain_inputs:
            dL_dpsi2 = beta[:, None] * dL_dpsi2_beta[None, :, :]
        else:
            dL_dpsi1 += 2.*np.dot(dL_dpsi2_beta, (psi1 * beta).T).T
            dL_dpsi2 = None
    else:
        dL_dpsi2 = beta * dL_dpsi2_beta
        if not uncertain_inputs:
            # subsume back into psi1 (==Kmn)
            dL_dpsi1 += 2.*np.dot(psi1, dL_dpsi2)
            dL_dpsi2 = None
    return dL_dpsi0, dL_dpsi1, dL_dpsi2


def _compute_dL_dR(likelihood, het_noise, uncertain_inputs, LB, _LBi_Lmi_psi1Vf, DBi_plus_BiPBi, Lm, A, psi0, psi1, beta, data_fit, num_data, output_dim, trYYT, Y, VVT_factr=None):
    # the partial derivative vector for the likelihood
    if likelihood.size == 0:
        # save computation here.
        dL_dR = None
    elif het_noise:
        if uncertain_inputs:
            raise NotImplementedError("heteroscedatic derivates with uncertain inputs not implemented")
        else:
            #from ...util.linalg import chol_inv
            #LBi = chol_inv(LB)
            LBi, _ = dtrtrs(LB,np.eye(LB.shape[0]))

            Lmi_psi1, nil = dtrtrs(Lm, psi1.T, lower=1, trans=0)
            _LBi_Lmi_psi1, _ = dtrtrs(LB, Lmi_psi1, lower=1, trans=0)
            dL_dR = -0.5 * beta + 0.5 * VVT_factr**2
            dL_dR += 0.5 * output_dim * (psi0 - np.sum(Lmi_psi1**2,0))[:,None] * beta**2

            dL_dR += 0.5*np.sum(mdot(LBi.T,LBi,Lmi_psi1)*Lmi_psi1,0)[:,None]*beta**2

            dL_dR += -np.dot(_LBi_Lmi_psi1Vf.T,_LBi_Lmi_psi1).T * Y * beta**2
            dL_dR += 0.5*np.dot(_LBi_Lmi_psi1Vf.T,_LBi_Lmi_psi1).T**2 * beta**2
    else:
        # likelihood is not heteroscedatic
        dL_dR = -0.5 * num_data * output_dim * beta + 0.5 * trYYT * beta ** 2
        dL_dR += 0.5 * output_dim * (psi0.sum() * beta ** 2 - np.trace(A) * beta)
        dL_dR += beta * (0.5 * np.sum(A * DBi_plus_BiPBi) - data_fit)
    return dL_dR

def _compute_log_marginal_likelihood(likelihood, num_data, output_dim, beta, het_noise, psi0, A, LB, trYYT, data_fit, Y):
    #compute log marginal likelihood
    if het_noise:
        lik_1 = -0.5 * num_data * output_dim * np.log(2. * np.pi) + 0.5 * output_dim * np.sum(np.log(beta)) - 0.5 * np.sum(beta.ravel() * np.square(Y).sum(axis=-1))
        lik_2 = -0.5 * output_dim * (np.sum(beta.flatten() * psi0) - np.trace(A))
    else:
        lik_1 = -0.5 * num_data * output_dim * (np.log(2. * np.pi) - np.log(beta)) - 0.5 * beta * trYYT
        lik_2 = -0.5 * output_dim * (np.sum(beta * psi0) - np.trace(A))
    lik_3 = -output_dim * (np.sum(np.log(np.diag(LB))))
    lik_4 = 0.5 * data_fit
    log_marginal = lik_1 + lik_2 + lik_3 + lik_4
    return log_marginal



# Copyright (c) 2012, GPy authors (see AUTHORS.txt).
# Licensed under the BSD 3-clause license (see LICENSE.txt)
import numpy as np
from scipy import special
from scipy.special import erfcx
import sys
epsilon = sys.float_info.epsilon
lim_val = -np.log(epsilon) 

def logisticln(x): # pragma: no cover
    return np.where(x<lim_val, np.where(x>-lim_val, -np.log(1+np.exp(-x)), -x), -np.log(1+epsilon))

def logistic(x): # pragma: no cover
    return special.expit(x)
    #return np.where(x<lim_val, np.where(x>-lim_val, 1/(1+np.exp(-x)), epsilon/(epsilon+1)), 1/(1+epsilon))

def normcdf(x): # pragma: no cover
    return special.ndtr(x)
    #g=0.5*erfc(-x/np.sqrt(2))
    #return np.where(g==0, epsilon, np.where(g==1, 1-epsilon, g)) 

def normcdfln(x): # pragma: no cover
    return special.log_ndtr(x)
    #return np.where(x < 0, -.5*x*x + np.log(.5) + np.log(erfcx(-x/np.sqrt(2))), np.log(normcdf(x)))

def clip_exp(x): # pragma: no cover
    return np.where(x<lim_val, np.where(x>-lim_val, np.exp(x), epsilon), 1/epsilon)

def differfln(x0, x1): # pragma: no cover
    # this is a, hopefully!, a numerically more stable variant of log(erf(x0)-erf(x1)) = log(erfc(x1)-erfc(x0)).
    return np.where(x0>x1, -x1*x1 + np.log(erfcx(x1)-np.exp(-x0**2+x1**2)*erfcx(x0)), -x0*x0 + np.log(np.exp(-x1**2+x0**2)*erfcx(x1) - erfcx(x0)))



# Copyright (c) 2014, Max Zwiessele, GPy Authors
# Licensed under the BSD 3-clause license (see LICENSE.txt)
import unittest
import sys

def deepTest(reason):
    if reason:
        return lambda x:x
    return unittest.skip("Not deep scanning, enable deepscan by adding 'deep' argument to unittest call")



# Copyright (c) 2012, GPy authors (see AUTHORS.txt).
# Licensed under the BSD 3-clause license (see LICENSE.txt)


from .kern import Kern
import numpy as np
from ...core.parameterization import Param
from paramz.transformations import Logexp
from paramz.caching import Cache_this

class Static(Kern):
    def __init__(self, input_dim, variance, active_dims, name):
        super(Static, self).__init__(input_dim, active_dims, name)
        self.variance = Param('variance', variance, Logexp())
        self.link_parameters(self.variance)

    def _save_to_input_dict(self):
        input_dict = super(Static, self)._save_to_input_dict()
        input_dict["variance"] =  self.variance.values.tolist()
        return input_dict

    def Kdiag(self, X):
        ret = np.empty((X.shape[0],), dtype=np.float64)
        ret[:] = self.variance
        return ret

    def gradients_X(self, dL_dK, X, X2=None):
        return np.zeros(X.shape)

    def gradients_X_diag(self, dL_dKdiag, X):
        return np.zeros(X.shape)

    def gradients_XX(self, dL_dK, X, X2=None):
        if X2 is None:
            X2 = X
        return np.zeros((X.shape[0], X2.shape[0], X.shape[1], X.shape[1]), dtype=np.float64)

    def gradients_XX_diag(self, dL_dKdiag, X, cov=False):
        return np.zeros((X.shape[0], X.shape[1], X.shape[1]), dtype=np.float64)

    def gradients_Z_expectations(self, dL_dpsi0, dL_dpsi1, dL_dpsi2, Z, variational_posterior):
        return np.zeros(Z.shape)

    def gradients_qX_expectations(self, dL_dpsi0, dL_dpsi1, dL_dpsi2, Z, variational_posterior):
        return np.zeros(variational_posterior.shape), np.zeros(variational_posterior.shape)

    def psi0(self, Z, variational_posterior):
        return self.Kdiag(variational_posterior.mean)

    def psi1(self, Z, variational_posterior):
        return self.K(variational_posterior.mean, Z)

    def psi2(self, Z, variational_posterior):
        K = self.K(variational_posterior.mean, Z)
        return np.einsum('ij,ik->jk',K,K) #K[:,:,None]*K[:,None,:] # NB. more efficient implementations on inherriting classes

    def input_sensitivity(self, summarize=True):
        if summarize:
            return super(Static, self).input_sensitivity(summarize=summarize)
        else:
            return np.ones(self.input_dim) * self.variance

class White(Static):
    def __init__(self, input_dim, variance=1., active_dims=None, name='white'):
        super(White, self).__init__(input_dim, variance, active_dims, name)

    def to_dict(self):
        input_dict = super(White, self)._save_to_input_dict()
        input_dict["class"] = "GPy.kern.White"
        return input_dict
    
    def K(self, X, X2=None):
        if X2 is None:
            return np.eye(X.shape[0])*self.variance
        else:
            return np.zeros((X.shape[0], X2.shape[0]))

    def psi2(self, Z, variational_posterior):
        return np.zeros((Z.shape[0], Z.shape[0]), dtype=np.float64)

    def psi2n(self, Z, variational_posterior):
        return np.zeros((1, Z.shape[0], Z.shape[0]), dtype=np.float64)

    def update_gradients_full(self, dL_dK, X, X2=None):
        if X2 is None:
            self.variance.gradient = np.trace(dL_dK)
        else:
            self.variance.gradient = 0.

    def update_gradients_diag(self, dL_dKdiag, X):
        self.variance.gradient = dL_dKdiag.sum()

    def update_gradients_expectations(self, dL_dpsi0, dL_dpsi1, dL_dpsi2, Z, variational_posterior):
        self.variance.gradient = dL_dpsi0.sum()

class WhiteHeteroscedastic(Static):
    def __init__(self, input_dim, num_data, variance=1., active_dims=None, name='white_hetero'):
        """
        A heteroscedastic White kernel (nugget/noise).
        It defines one variance (nugget) per input sample.

        Prediction excludes any noise learnt by this Kernel, so be careful using this kernel.

        You can plot the errors learnt by this kernel by something similar as:
        plt.errorbar(m.X, m.Y, yerr=2*np.sqrt(m.kern.white.variance))
        """
        super(Static, self).__init__(input_dim, active_dims, name)
        self.variance = Param('variance', np.ones(num_data) * variance, Logexp())
        self.link_parameters(self.variance)
    def to_dict(self):
        input_dict = super(WhiteHeteroscedastic, self)._save_to_input_dict()
        input_dict["class"] = "GPy.kern.WhiteHeteroscedastic"
        return input_dict

    def Kdiag(self, X):
        if X.shape[0] == self.variance.shape[0]:
            # If the input has the same number of samples as
            # the number of variances, we return the variances
            return self.variance
        return 0.

    def K(self, X, X2=None):
        if X2 is None and X.shape[0] == self.variance.shape[0]:
            return np.eye(X.shape[0]) * self.variance
        else:
            return 0.

    def psi2(self, Z, variational_posterior):
        return np.zeros((Z.shape[0], Z.shape[0]), dtype=np.float64)

    def psi2n(self, Z, variational_posterior):
        return np.zeros((1, Z.shape[0], Z.shape[0]), dtype=np.float64)

    def update_gradients_full(self, dL_dK, X, X2=None):
        if X2 is None:
            self.variance.gradient = np.diagonal(dL_dK)
        else:
            self.variance.gradient = 0.

    def update_gradients_diag(self, dL_dKdiag, X):
        self.variance.gradient = dL_dKdiag

    def update_gradients_expectations(self, dL_dpsi0, dL_dpsi1, dL_dpsi2, Z, variational_posterior):
        self.variance.gradient = dL_dpsi0

class Bias(Static):
    def __init__(self, input_dim, variance=1., active_dims=None, name='bias'):
        super(Bias, self).__init__(input_dim, variance, active_dims, name)

    def to_dict(self):
        input_dict = super(Bias, self)._save_to_input_dict()
        input_dict["class"] = "GPy.kern.Bias"
        return input_dict

    @staticmethod
    def _build_from_input_dict(kernel_class, input_dict):
        useGPU = input_dict.pop('useGPU', None)
        return Bias(**input_dict)

    def K(self, X, X2=None):
        shape = (X.shape[0], X.shape[0] if X2 is None else X2.shape[0])
        return np.full(shape, self.variance, dtype=np.float64)

    def update_gradients_full(self, dL_dK, X, X2=None):
        self.variance.gradient = dL_dK.sum()

    def update_gradients_diag(self, dL_dKdiag, X):
        self.variance.gradient = dL_dKdiag.sum()

    def psi2(self, Z, variational_posterior):
        return np.full((Z.shape[0], Z.shape[0]), self.variance*self.variance*variational_posterior.shape[0], dtype=np.float64)

    def psi2n(self, Z, variational_posterior):
        ret = np.empty((variational_posterior.mean.shape[0], Z.shape[0], Z.shape[0]), dtype=np.float64)
        ret[:] = self.variance*self.variance
        return ret

    def update_gradients_expectations(self, dL_dpsi0, dL_dpsi1, dL_dpsi2, Z, variational_posterior):
        if dL_dpsi2.ndim == 2:
            self.variance.gradient = (dL_dpsi0.sum() + dL_dpsi1.sum()
                                    + 2.*self.variance*dL_dpsi2.sum()*variational_posterior.shape[0])
        else:
            self.variance.gradient = (dL_dpsi0.sum() + dL_dpsi1.sum()
                                    + 2.*self.variance*dL_dpsi2.sum())

class Fixed(Static):
    def __init__(self, input_dim, covariance_matrix, variance=1., active_dims=None, name='fixed'):
        """
        :param input_dim: the number of input dimensions
        :type input_dim: int
        :param variance: the variance of the kernel
        :type variance: float
        """
        super(Fixed, self).__init__(input_dim, variance, active_dims, name)
        self.fixed_K = covariance_matrix
    def K(self, X, X2):
        if X2 is None:
            return self.variance * self.fixed_K
        else:
            return np.zeros((X.shape[0], X2.shape[0]))

    def Kdiag(self, X):
        return self.variance * self.fixed_K.diagonal()

    def update_gradients_full(self, dL_dK, X, X2=None):
        if X2 is None:
            self.variance.gradient = np.einsum('ij,ij', dL_dK, self.fixed_K)
        else:
            self.variance.gradient = 0

    def update_gradients_diag(self, dL_dKdiag, X):
        self.variance.gradient = np.einsum('i,i', dL_dKdiag, np.diagonal(self.fixed_K))

    def psi2(self, Z, variational_posterior):
        return np.zeros((Z.shape[0], Z.shape[0]), dtype=np.float64)

    def psi2n(self, Z, variational_posterior):
        return np.zeros((1, Z.shape[0], Z.shape[0]), dtype=np.float64)

    def update_gradients_expectations(self, dL_dpsi0, dL_dpsi1, dL_dpsi2, Z, variational_posterior):
        self.variance.gradient = dL_dpsi0.sum()

class Precomputed(Fixed):
    def __init__(self, input_dim, covariance_matrix, variance=1., active_dims=None, name='precomputed'):
        """
        Class for precomputed kernels, indexed by columns in X

        Usage example:

        import numpy as np
        from GPy.models import GPClassification
        from GPy.kern import Precomputed
        from sklearn.cross_validation import LeaveOneOut

        n = 10
        d = 100
        X = np.arange(n).reshape((n,1))         # column vector of indices
        y = 2*np.random.binomial(1,0.5,(n,1))-1
        X0 = np.random.randn(n,d)
        k = np.dot(X0,X0.T)
        kern = Precomputed(1,k)                 # k is a n x n covariance matrix

        cv = LeaveOneOut(n)
        ypred = y.copy()
        for train, test in cv:
            m = GPClassification(X[train], y[train], kernel=kern)
            m.optimize()
            ypred[test] = 2*(m.predict(X[test])[0]>0.5)-1

        :param input_dim: the number of input dimensions
        :type input_dim: int
        :param variance: the variance of the kernel
        :type variance: float
        """
        assert input_dim==1, "Precomputed only implemented in one dimension. Use multiple Precomputed kernels to have more dimensions by making use of active_dims"
        super(Precomputed, self).__init__(input_dim, covariance_matrix, variance, active_dims, name)

    @Cache_this(limit=2)
    def _index(self, X, X2):
        if X2 is None:
            i1 = i2 = X.astype('int').flat
        else:
            i1, i2 = X.astype('int').flat, X2.astype('int').flat
        return self.fixed_K[i1,:][:,i2]

    def K(self, X, X2=None):
        return self.variance * self._index(X, X2)

    def Kdiag(self, X):
        return self.variance * self._index(X,None).diagonal()

    def update_gradients_full(self, dL_dK, X, X2=None):
        self.variance.gradient = np.einsum('ij,ij', dL_dK, self._index(X, X2))

    def update_gradients_diag(self, dL_dKdiag, X):
        self.variance.gradient = np.einsum('i,ii', dL_dKdiag, self._index(X, None))



# Copyright (c) 2012, GPy authors (see AUTHORS.txt).
# Licensed under the BSD 3-clause license (see LICENSE.txt)

import numpy as np

def view(A, offset=0):
    """
    Get a view on the diagonal elements of a 2D array.

    This is actually a view (!) on the diagonal of the array, so you can
    in-place adjust the view.

    :param :class:`ndarray` A: 2 dimensional numpy array
    :param int offset: view offset to give back (negative entries allowed)
    :rtype: :class:`ndarray` view of diag(A)

    >>> import numpy as np
    >>> X = np.arange(9).reshape(3,3)
    >>> view(X)
    array([0, 4, 8])
    >>> d = view(X)
    >>> d += 2
    >>> view(X)
    array([ 2,  6, 10])
    >>> view(X, offset=-1)
    array([3, 7])
    >>> subtract(X, 3, offset=-1)
    array([[ 2,  1,  2],
           [ 0,  6,  5],
           [ 6,  4, 10]])
    """
    from numpy.lib.stride_tricks import as_strided
    assert A.ndim == 2, "only implemented for 2 dimensions"
    assert A.shape[0] == A.shape[1], "attempting to get the view of non-square matrix?!"
    if offset > 0:
        return as_strided(A[0, offset:], shape=(A.shape[0] - offset, ), strides=((A.shape[0]+1)*A.itemsize, ))
    elif offset < 0:
        return as_strided(A[-offset:, 0], shape=(A.shape[0] + offset, ), strides=((A.shape[0]+1)*A.itemsize, ))
    else:
        return as_strided(A, shape=(A.shape[0], ), strides=((A.shape[0]+1)*A.itemsize, ))

def offdiag_view(A, offset=0):
    from numpy.lib.stride_tricks import as_strided
    assert A.ndim == 2, "only implemented for 2 dimensions"
    Af = as_strided(A, shape=(A.size,), strides=(A.itemsize,))
    return as_strided(Af[(1+offset):], shape=(A.shape[0]-1, A.shape[1]), strides=(A.strides[0] + A.itemsize, A.strides[1]))

def _diag_ufunc(A,b,offset,func):
    b = np.squeeze(b)
    assert b.ndim <= 1, "only implemented for one dimensional arrays"
    dA = view(A, offset); func(dA,b,dA)
    return A

def times(A, b, offset=0):
    """
    Times the view of A with b in place (!).
    Returns modified A
    Broadcasting is allowed, thus b can be scalar.

    if offset is not zero, make sure b is of right shape!

    :param ndarray A: 2 dimensional array
    :param ndarray-like b: either one dimensional or scalar
    :param int offset: same as in view.
    :rtype: view of A, which is adjusted inplace
    """
    return _diag_ufunc(A, b, offset, np.multiply)
multiply = times

def divide(A, b, offset=0):
    """
    Divide the view of A by b in place (!).
    Returns modified A
    Broadcasting is allowed, thus b can be scalar.

    if offset is not zero, make sure b is of right shape!

    :param ndarray A: 2 dimensional array
    :param ndarray-like b: either one dimensional or scalar
    :param int offset: same as in view.
    :rtype: view of A, which is adjusted inplace
    """
    return _diag_ufunc(A, b, offset, np.divide)

def add(A, b, offset=0):
    """
    Add b to the view of A in place (!).
    Returns modified A.
    Broadcasting is allowed, thus b can be scalar.

    if offset is not zero, make sure b is of right shape!

    :param ndarray A: 2 dimensional array
    :param ndarray-like b: either one dimensional or scalar
    :param int offset: same as in view.
    :rtype: view of A, which is adjusted inplace
    """
    return _diag_ufunc(A, b, offset, np.add)

def subtract(A, b, offset=0):
    """
    Subtract b from the view of A in place (!).
    Returns modified A.
    Broadcasting is allowed, thus b can be scalar.

    if offset is not zero, make sure b is of right shape!

    :param ndarray A: 2 dimensional array
    :param ndarray-like b: either one dimensional or scalar
    :param int offset: same as in view.
    :rtype: view of A, which is adjusted inplace
    """
    return _diag_ufunc(A, b, offset, np.subtract)

if __name__ == '__main__':
    import doctest
    doctest.testmod()
